"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest

import salt.modules.mac_group as mac_group
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, patch

grp = pytest.importorskip("grp", reason="Missing required library 'grp'")


@pytest.fixture
def configure_loader_modules():
    return {mac_group: {}}


def test_add_group_exists():
    """
    Tests if the group already exists or not
    """
    mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
    with patch("salt.modules.mac_group.info", MagicMock(return_value=mock_group)):
        pytest.raises(CommandExecutionError, mac_group.add, "test")


def test_add_whitespace():
    """
    Tests if there is whitespace in the group name
    """
    with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
        pytest.raises(SaltInvocationError, mac_group.add, "white space")


def test_add_underscore():
    """
    Tests if the group name starts with an underscore or not
    """
    with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
        pytest.raises(SaltInvocationError, mac_group.add, "_Test")


def test_add_gid_int():
    """
    Tests if the gid is an int or not
    """
    with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
        pytest.raises(SaltInvocationError, mac_group.add, "foo", "foo")


def test_add_gid_exists():
    """
    Tests if the gid is already in use or not
    """
    with patch("salt.modules.mac_group.info", MagicMock(return_value={})), patch(
        "salt.modules.mac_group._list_gids", MagicMock(return_value=["3456"])
    ):
        pytest.raises(CommandExecutionError, mac_group.add, "foo", 3456)


def test_add():
    """
    Tests if specified group was added
    """
    mock_ret = MagicMock(return_value=0)
    with patch.dict(mac_group.__salt__, {"cmd.retcode": mock_ret}), patch(
        "salt.modules.mac_group.info", MagicMock(return_value={})
    ), patch("salt.modules.mac_group._list_gids", MagicMock(return_value=[])):
        assert mac_group.add("test", 500)


def test_delete_whitespace():
    """
    Tests if there is whitespace in the group name
    """
    pytest.raises(SaltInvocationError, mac_group.delete, "white space")


def test_delete_underscore():
    """
    Tests if the group name starts with an underscore or not
    """
    pytest.raises(SaltInvocationError, mac_group.delete, "_Test")


def test_delete_group_exists():
    """
    Tests if the group to be deleted exists or not
    """
    with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
        assert mac_group.delete("test")


def test_delete():
    """
    Tests if the specified group was deleted
    """
    mock_ret = MagicMock(return_value=0)
    mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
    with patch.dict(mac_group.__salt__, {"cmd.retcode": mock_ret}), patch(
        "salt.modules.mac_group.info", MagicMock(return_value=mock_group)
    ):
        assert mac_group.delete("test")


def test_info_whitespace():
    """
    Tests if there is whitespace in the group name
    """
    pytest.raises(SaltInvocationError, mac_group.info, "white space")


def test_info():
    """
    Tests the return of group information
    """
    mock_getgrall = [grp.struct_group(("foo", "*", 20, ["test"]))]
    with patch("grp.getgrall", MagicMock(return_value=mock_getgrall)):
        ret = {"passwd": "*", "gid": 20, "name": "foo", "members": ["test"]}
        assert mac_group.info("foo") == ret


def test_format_info():
    """
    Tests the formatting of returned group information
    """
    data = grp.struct_group(("wheel", "*", 0, ["root"]))
    ret = {"passwd": "*", "gid": 0, "name": "wheel", "members": ["root"]}
    assert mac_group._format_info(data) == ret


def test_getent():
    """
    Tests the return of information on all groups
    """
    mock_getgrall = [grp.struct_group(("foo", "*", 20, ["test"]))]
    with patch("grp.getgrall", MagicMock(return_value=mock_getgrall)):
        ret = [{"passwd": "*", "gid": 20, "name": "foo", "members": ["test"]}]
        assert mac_group.getent() == ret


def test_chgid_gid_int():
    """
    Tests if gid is an integer or not
    """
    pytest.raises(SaltInvocationError, mac_group.chgid, "foo", "foo")


def test_chgid_group_exists():
    """
    Tests if the group id exists or not
    """
    mock_pre_gid = MagicMock(return_value="")
    with patch.dict(mac_group.__salt__, {"file.group_to_gid": mock_pre_gid}), patch(
        "salt.modules.mac_group.info", MagicMock(return_value={})
    ):
        pytest.raises(CommandExecutionError, mac_group.chgid, "foo", 4376)


def test_chgid_gid_same():
    """
    Tests if the group id is the same as argument
    """
    mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
    mock_pre_gid = MagicMock(return_value=0)
    with patch.dict(mac_group.__salt__, {"file.group_to_gid": mock_pre_gid}), patch(
        "salt.modules.mac_group.info", MagicMock(return_value=mock_group)
    ):
        assert mac_group.chgid("test", 0)


def test_chgid():
    """
    Tests the gid for a named group was changed
    """
    mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
    mock_pre_gid = MagicMock(return_value=0)
    mock_ret = MagicMock(return_value=0)
    with patch.dict(
        mac_group.__salt__, {"file.group_to_gid": mock_pre_gid}
    ), patch.dict(mac_group.__salt__, {"cmd.retcode": mock_ret}), patch(
        "salt.modules.mac_group.info", MagicMock(return_value=mock_group)
    ):
        assert mac_group.chgid("test", 500)



import logging

import pytest

import salt.modules.status as status
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():

    return {
        status: {
            "__grains__": {
                "ip4_interfaces": {
                    "en0": ["129.40.94.58"],
                    "en1": ["172.24.94.58"],
                    "lo0": ["127.0.0.1"],
                },
                "ip6_interfaces": {"en0": [], "en1": [], "lo0": ["1"]},
                "kernel": "AIX",
                "osarch": "PowerPC_POWER8",
                "os": "AIX",
                "os_family": "AIX",
                "osmajorrelease": 7,
            },
        },
    }


def test_netdev():
    """
    Test status.netdev for AIX

    :return:
    """
    # Output from netstat -i -n -I <en0|en1|lo0> -f inet
    netstat_inet4_en0 = """Name   Mtu   Network     Address                 Ipkts     Ierrs        Opkts     Oerrs  Coll
en0    1500  link#2      fa.41.f5.e9.bd.20  1523125     0   759364     0     0
en0    1500  129.40.94.5 129.40.94.58      1523125     0   759364     0     0
"""

    netstat_inet4_en1 = """Name   Mtu   Network     Address                 Ipkts     Ierrs        Opkts     Oerrs  Coll
en1    1500  link#3      fa.41.f5.e9.bd.21     1089     0      402     0     0
en1    1500  172.24.94.5 172.24.94.58         1089     0      402     0     0
"""

    netstat_inet4_lo0 = """Name   Mtu   Network     Address                 Ipkts     Ierrs        Opkts     Oerrs  Coll
lo0    16896 link#1                          25568     0    25568     0     0
lo0    16896 127         127.0.0.1           25568     0    25568     0     0
"""

    # Output from netstat -i -n -I <en0|en1|lo0> -f inet6
    netstat_inet6_en0 = """Name   Mtu   Network     Address                 Ipkts     Ierrs        Opkts     Oerrs  Coll
en0    1500  link#2      fa.41.f5.e9.bd.20  1523160     0   759397     0     0
"""

    netstat_inet6_en1 = """Name   Mtu   Network     Address                 Ipkts     Ierrs        Opkts     Oerrs  Coll
en1    1500  link#3      fa.41.f5.e9.bd.21     1089     0      402     0     0
"""

    netstat_inet6_lo0 = """Name   Mtu   Network     Address                 Ipkts     Ierrs        Opkts     Oerrs  Coll
lo0    16896 link#1                          25611     0    25611     0     0
lo0    16896 ::1%1                           25611     0    25611     0     0
"""

    # allow en0, en1 and lo0 for ipv4 and ipv6
    netstats_out = MagicMock(
        side_effect=[
            netstat_inet4_en0,
            netstat_inet6_en0,
            netstat_inet4_en1,
            netstat_inet6_en1,
            netstat_inet4_lo0,
            netstat_inet6_lo0,
            netstat_inet4_en0,
            netstat_inet6_en0,
            netstat_inet4_en1,
            netstat_inet6_en1,
            netstat_inet4_lo0,
            netstat_inet6_lo0,
        ]
    )

    with patch.dict(
        status.__grains__,
        {
            "osarch": "PowerPC_POWER8",
            "ip4_interfaces": {
                "en0": ["129.40.94.58"],
                "en1": ["172.24.94.58"],
                "lo0": ["127.0.0.1"],
            },
            "ip6_interfaces": {
                "en0": [],
                "en1": [],
                "lo0": ["::1"],
            },
            "kernel": "AIX",
        },
    ), patch.dict(status.__salt__, {"cmd.run": netstats_out}):
        netdev_out = status.netdev()
        assert netstats_out.call_count == 12
        netstats_out.assert_any_call("netstat -i -n -I en0 -f inet")
        netstats_out.assert_any_call("netstat -i -n -I en1 -f inet")
        netstats_out.assert_any_call("netstat -i -n -I lo0 -f inet")
        netstats_out.assert_any_call("netstat -i -n -I en0 -f inet6")
        netstats_out.assert_any_call("netstat -i -n -I en1 -f inet6")
        netstats_out.assert_any_call("netstat -i -n -I lo0 -f inet6")
        expected = {
            "en0": [
                {
                    "ipv4": {
                        "Mtu": "1500",
                        "Network": "129.40.94.5",
                        "Address": "129.40.94.58",
                        "Ipkts": "1523125",
                        "Ierrs": "0",
                        "Opkts": "759364",
                        "Oerrs": "0",
                        "Coll": "0",
                    }
                }
            ],
            "en1": [
                {
                    "ipv4": {
                        "Mtu": "1500",
                        "Network": "172.24.94.5",
                        "Address": "172.24.94.58",
                        "Ipkts": "1089",
                        "Ierrs": "0",
                        "Opkts": "402",
                        "Oerrs": "0",
                        "Coll": "0",
                    }
                }
            ],
            "lo0": [
                {
                    "ipv4": {
                        "Mtu": "16896",
                        "Network": "127",
                        "Address": "127.0.0.1",
                        "Ipkts": "25568",
                        "Ierrs": "0",
                        "Opkts": "25568",
                        "Oerrs": "0",
                        "Coll": "0",
                    }
                },
                {
                    "ipv6": {
                        "Mtu": "16896",
                        "Network": "::1%1",
                        "Address": "25611",
                        "Ipkts": "0",
                        "Ierrs": "25611",
                        "Opkts": "0",
                        "Oerrs": "0",
                    }
                },
            ],
        }
        assert netdev_out == expected



import pytest


@pytest.fixture
def blackout_minion(salt_minion, salt_master, salt_run_cli):
    with pytest.helpers.temp_file(
        "top.sls",
        """base:\n  '*':\n    - fnord""",
        salt_master.pillar_tree.base.temp_file,
    ), pytest.helpers.temp_file(
        "fnord.sls", """minion_blackout: True""", salt_master.pillar_tree.base.paths[0]
    ):
        op = salt_run_cli.run("saltutil.sync_all")
        yield op


@pytest.fixture
def non_blackout_minion(salt_minion, salt_master, salt_call_cli):
    with pytest.helpers.temp_file(
        "top.sls",
        """base:\n  '*':\n    - fnord""",
        salt_master.pillar_tree.base.paths[0],
    ), pytest.helpers.temp_file(
        "fnord.sls", """minion_blackout: False""", salt_master.pillar_tree.base.paths[0]
    ):
        op = salt_call_cli.run("saltutil.refresh_pillar")
        # Refresh pillar appears to return too quickly. This call ensures that
        # we wait until the pillar refresh gets the correct value in there.
        op = salt_call_cli.run("pillar.get", "minion_blackout")
        yield op


def test_when_manage_versions_runs_and_the_minion_errors_out_it_should_not_(
    salt_run_cli, blackout_minion
):
    op = salt_run_cli.run("manage.versions")
    assert "Unable to parse version string" not in op.stdout


def test_when_manage_versions_runs_and_the_minion_does_not_error_it_should_have_version_info(
    salt_run_cli, non_blackout_minion
):
    op = salt_run_cli.run("manage.versions")
    result = op.stdout
    assert "Up to date" in result



"""
Test module for syslog_ng state
"""

import os
import re
import tempfile

import salt.modules.syslog_ng as syslog_ng_module
import salt.states.syslog_ng as syslog_ng
import salt.utils.files
import salt.utils.yaml
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

SOURCE_1_CONFIG = {
    "id": "s_tail",
    "config": """
        source:
            - file:
              - '"/var/log/apache/access.log"'
              - follow_freq : 1
              - flags:
                - no-parse
                - validate-utf8
        """,
}

SOURCE_1_EXPECTED = """
source s_tail {
   file(
         "/var/log/apache/access.log",
         follow_freq(1),
         flags(no-parse, validate-utf8)
   );
};
"""

SOURCE_2_CONFIG = {
    "id": "s_gsoc2014",
    "config": """
        source:
          - tcp:
            - ip: '"0.0.0.0"'
            - port: 1234
            - flags: no-parse
        """,
}

SOURCE_2_EXPECTED = """
source s_gsoc2014 {
   tcp(
         ip("0.0.0.0"),
         port(1234),
         flags(no-parse)
   );
};"""

FILTER_1_CONFIG = {
    "id": "f_json",
    "config": """
        filter:
          - match:
            - '"@json:"'
        """,
}

FILTER_1_EXPECTED = """
    filter f_json {
       match(
             "@json:"
       );
    };
    """

TEMPLATE_1_CONFIG = {
    "id": "t_demo_filetemplate",
    "config": """
        template:
          - template:
            - '"$ISODATE $HOST $MSG\n"'
          - template_escape:
            - "no"
        """,
}

TEMPLATE_1_EXPECTED = """
    template t_demo_filetemplate {
       template(
             "$ISODATE $HOST $MSG "
       );
       template_escape(
             no
       );
    };
    """

REWRITE_1_CONFIG = {
    "id": "r_set_message_to_MESSAGE",
    "config": """
        rewrite:
          - set:
            - '"${.json.message}"'
            - value : '"$MESSAGE"'
        """,
}

REWRITE_1_EXPECTED = """
    rewrite r_set_message_to_MESSAGE {
       set(
             "${.json.message}",
             value("$MESSAGE")
       );
    };
    """

LOG_1_CONFIG = {
    "id": "l_gsoc2014",
    "config": """
        log:
          - source: s_gsoc2014
          - junction:
            - channel:
              - filter: f_json
              - parser: p_json
              - rewrite: r_set_json_tag
              - rewrite: r_set_message_to_MESSAGE
              - destination:
                - file:
                  - '"/tmp/json-input.log"'
                  - template: t_gsoc2014
              - flags: final
            - channel:
              - filter: f_not_json
              - parser:
                - syslog-parser: []
              - rewrite: r_set_syslog_tag
              - flags: final
          - destination:
            - file:
              - '"/tmp/all.log"'
              - template: t_gsoc2014
        """,
}

LOG_1_EXPECTED = """
    log {
       source(s_gsoc2014);
       junction {
          channel {
             filter(f_json);
             parser(p_json);
             rewrite(r_set_json_tag);
             rewrite(r_set_message_to_MESSAGE);
             destination {
                file(
                      "/tmp/json-input.log",
                      template(t_gsoc2014)
                );
             };
             flags(final);
          };
          channel {
             filter(f_not_json);
             parser {
                syslog-parser(

                );
             };
             rewrite(r_set_syslog_tag);
             flags(final);
          };
       };
       destination {
          file(
                "/tmp/all.log",
                template(t_gsoc2014)
          );
       };
    };
    """

OPTIONS_1_CONFIG = {
    "id": "global_options",
    "config": """
        options:
          - time_reap: 30
          - mark_freq: 10
          - keep_hostname: "yes"
        """,
}

OPTIONS_1_EXPECTED = """
    options {
        time_reap(30);
        mark_freq(10);
        keep_hostname(yes);
    };
    """

SHORT_FORM_CONFIG = {
    "id": "source.s_gsoc",
    "config": """
          - tcp:
            - ip: '"0.0.0.0"'
            - port: 1234
            - flags: no-parse
        """,
}

SHORT_FORM_EXPECTED = """
    source s_gsoc {
        tcp(
            ip(
                "0.0.0.0"
            ),
            port(
                1234
            ),
            flags(
              no-parse
            )
        );
    };
    """

GIVEN_CONFIG = {
    "id": "config.some_name",
    "config": """
               source s_gsoc {
                  tcp(
                      ip(
                          "0.0.0.0"
                      ),
                      port(
                          1234
                      ),
                      flags(
                        no-parse
                      )
                  );
               };
        """,
}

_SALT_VAR_WITH_MODULE_METHODS = {
    "syslog_ng.config": syslog_ng_module.config,
    "syslog_ng.start": syslog_ng_module.start,
    "syslog_ng.reload": syslog_ng_module.reload_,
    "syslog_ng.stop": syslog_ng_module.stop,
    "syslog_ng.write_version": syslog_ng_module.write_version,
    "syslog_ng.write_config": syslog_ng_module.write_config,
}


def remove_whitespaces(source):
    return re.sub(r"\s+", "", source.strip())


class SyslogNGTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {syslog_ng: {}, syslog_ng_module: {"__opts__": {"test": False}}}

    def test_generate_source_config(self):
        self._config_generator_template(SOURCE_1_CONFIG, SOURCE_1_EXPECTED)

    def test_generate_log_config(self):
        self._config_generator_template(LOG_1_CONFIG, LOG_1_EXPECTED)

    def test_generate_tcp_source_config(self):
        self._config_generator_template(SOURCE_2_CONFIG, SOURCE_2_EXPECTED)

    def test_generate_filter_config(self):
        self._config_generator_template(FILTER_1_CONFIG, FILTER_1_EXPECTED)

    def test_generate_template_config(self):
        self._config_generator_template(TEMPLATE_1_CONFIG, TEMPLATE_1_EXPECTED)

    def test_generate_rewrite_config(self):
        self._config_generator_template(REWRITE_1_CONFIG, REWRITE_1_EXPECTED)

    def test_generate_global_options_config(self):
        self._config_generator_template(OPTIONS_1_CONFIG, OPTIONS_1_EXPECTED)

    def test_generate_short_form_statement(self):
        self._config_generator_template(SHORT_FORM_CONFIG, SHORT_FORM_EXPECTED)

    def test_generate_given_config(self):
        self._config_generator_template(GIVEN_CONFIG, SHORT_FORM_EXPECTED)

    def _config_generator_template(self, yaml_input, expected):
        parsed_yaml_config = salt.utils.data.decode(
            salt.utils.yaml.safe_load(yaml_input["config"])
        )
        id = yaml_input["id"]

        with patch.dict(syslog_ng.__salt__, _SALT_VAR_WITH_MODULE_METHODS):
            got = syslog_ng.config(id, config=parsed_yaml_config, write=False)
            config = got["changes"]["new"]
            self.assertEqual(remove_whitespaces(expected), remove_whitespaces(config))
            self.assertEqual(False, got["result"])

    def test_write_config(self):
        yaml_inputs = (
            SOURCE_2_CONFIG,
            SOURCE_1_CONFIG,
            FILTER_1_CONFIG,
            TEMPLATE_1_CONFIG,
            REWRITE_1_CONFIG,
            LOG_1_CONFIG,
        )
        expected_outputs = (
            SOURCE_2_EXPECTED,
            SOURCE_1_EXPECTED,
            FILTER_1_EXPECTED,
            TEMPLATE_1_EXPECTED,
            REWRITE_1_EXPECTED,
            LOG_1_EXPECTED,
        )
        config_file_fd, config_file_name = tempfile.mkstemp()
        os.close(config_file_fd)

        with patch.dict(syslog_ng.__salt__, _SALT_VAR_WITH_MODULE_METHODS):
            syslog_ng_module.set_config_file(config_file_name)
            syslog_ng_module.write_version("3.6")
            syslog_ng_module.write_config(config='@include "scl.conf"')

            for i in yaml_inputs:
                parsed_yaml_config = salt.utils.data.decode(
                    salt.utils.yaml.safe_load(i["config"])
                )
                id = i["id"]
                got = syslog_ng.config(id, config=parsed_yaml_config, write=True)

            written_config = ""
            with salt.utils.files.fopen(config_file_name, "r") as f:
                written_config = salt.utils.stringutils.to_unicode(f.read())

            config_without_whitespaces = remove_whitespaces(written_config)
            for i in expected_outputs:
                without_whitespaces = remove_whitespaces(i)
                self.assertIn(without_whitespaces, config_without_whitespaces)

            syslog_ng_module.set_config_file("")
            os.remove(config_file_name)

    def test_started_state_generate_valid_cli_command(self):
        mock_func = MagicMock(return_value={"retcode": 0, "stdout": "", "pid": 1000})

        with patch.dict(syslog_ng.__salt__, _SALT_VAR_WITH_MODULE_METHODS):
            with patch.dict(syslog_ng_module.__salt__, {"cmd.run_all": mock_func}):
                got = syslog_ng.started(user="joe", group="users", enable_core=True)
                command = got["changes"]["new"]
                self.assertTrue(
                    command.endswith(
                        "syslog-ng --user=joe --group=users --enable-core"
                        " --cfgfile=/etc/syslog-ng.conf"
                    )
                )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
    :codeauthor: Jason Woods <devel@jasonwoods.me.uk>
"""

import pytest

import salt.states.selinux as selinux
from tests.support.mock import MagicMock, patch

pytestmark = [pytest.mark.skip_unless_on_linux]


@pytest.fixture
def configure_loader_modules():
    return {selinux: {}}


def test_mode():
    """
    Test to verifies the mode SELinux is running in,
    can be set to enforcing or permissive.
    """
    ret = {
        "name": "unknown",
        "changes": {},
        "result": False,
        "comment": "unknown is not an accepted mode",
    }
    assert selinux.mode("unknown") == ret

    mock_en = MagicMock(return_value="Enforcing")
    mock_pr = MagicMock(side_effect=["Permissive", "Enforcing"])
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.getenforce": mock_en,
            "selinux.getconfig": mock_en,
            "selinux.setenforce": mock_pr,
        },
    ):
        comt = "SELinux is already in Enforcing mode"
        ret = {"name": "Enforcing", "comment": comt, "result": True, "changes": {}}
        assert selinux.mode("Enforcing") == ret

        with patch.dict(selinux.__opts__, {"test": True}):
            comt = "SELinux mode is set to be changed to Permissive"
            ret = {
                "name": "Permissive",
                "comment": comt,
                "result": None,
                "changes": {"new": "Permissive", "old": "Enforcing"},
            }
            assert selinux.mode("Permissive") == ret

        with patch.dict(selinux.__opts__, {"test": False}):
            comt = "SELinux has been set to Permissive mode"
            ret = {
                "name": "Permissive",
                "comment": comt,
                "result": True,
                "changes": {"new": "Permissive", "old": "Enforcing"},
            }
            assert selinux.mode("Permissive") == ret

            comt = "Failed to set SELinux to Permissive mode"
            ret.update(
                {"name": "Permissive", "comment": comt, "result": False, "changes": {}}
            )
            assert selinux.mode("Permissive") == ret


def test_boolean():
    """
    Test to set up an SELinux boolean.
    """
    name = "samba_create_home_dirs"
    value = True
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    mock_en = MagicMock(return_value=[])
    with patch.dict(selinux.__salt__, {"selinux.list_sebool": mock_en}):
        comt = f"Boolean {name} is not available"
        ret.update({"comment": comt})
        assert selinux.boolean(name, value) == ret

    mock_bools = MagicMock(return_value={name: {"State": "on", "Default": "on"}})
    with patch.dict(selinux.__salt__, {"selinux.list_sebool": mock_bools}):
        comt = "None is not a valid value for the boolean"
        ret.update({"comment": comt})
        assert selinux.boolean(name, None) == ret

        comt = "Boolean is in the correct state"
        ret.update({"comment": comt, "result": True})
        assert selinux.boolean(name, value, True) == ret

        comt = "Boolean is in the correct state"
        ret.update({"comment": comt, "result": True})
        assert selinux.boolean(name, value) == ret

    mock_bools = MagicMock(return_value={name: {"State": "off", "Default": "on"}})
    mock = MagicMock(side_effect=[True, False])
    with patch.dict(
        selinux.__salt__,
        {"selinux.list_sebool": mock_bools, "selinux.setsebool": mock},
    ):
        with patch.dict(selinux.__opts__, {"test": True}):
            comt = "Boolean samba_create_home_dirs is set to be changed to on"
            ret.update({"comment": comt, "result": None})
            assert selinux.boolean(name, value) == ret

        with patch.dict(selinux.__opts__, {"test": False}):
            comt = "Boolean samba_create_home_dirs has been set to on"
            ret.update({"comment": comt, "result": True})
            ret.update({"changes": {"State": {"old": "off", "new": "on"}}})
            assert selinux.boolean(name, value) == ret

            comt = "Failed to set the boolean samba_create_home_dirs to on"
            ret.update({"comment": comt, "result": False})
            ret.update({"changes": {}})
            assert selinux.boolean(name, value) == ret


def test_port_policy_present():
    """
    Test to set up an SELinux port.
    """
    name = "tcp/8080"
    protocol = "tcp"
    port = "8080"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Test when already present with same sel_type
    mock_add = MagicMock(return_value={"retcode": 0})
    mock_modify = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(
        return_value={
            "sel_type": "http_cache_port_t",
            "protocol": "tcp",
            "port": "8080",
        }
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_add_policy": mock_add,
            "selinux.port_modify_policy": mock_modify,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": False}):
            comt = (
                f'SELinux policy for "{name}" already present '
                + 'with specified sel_type "http_cache_port_t", protocol "None" '
                + 'and port "None".'
            )
            ret.update({"comment": comt, "result": True})
            assert selinux.port_policy_present(name, "http_cache_port_t") == ret

            comt = (
                'SELinux policy for "name" already present '
                + f'with specified sel_type "http_cache_port_t", protocol "{protocol}" '
                + f'and port "{port}".'
            )
            ret.update({"comment": comt, "changes": {}, "result": True, "name": "name"})
            assert (
                selinux.port_policy_present("name", "http_cache_port_t", protocol, port)
                == ret
            )
            ret.update({"name": name})

    # Test adding new port policy
    mock_add = MagicMock(return_value={"retcode": 0})
    mock_modify = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(
        side_effect=[
            None,
            None,
            None,
            {"sel_type": "http_cache_port_t", "protocol": "tcp", "port": "8080"},
        ]
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_add_policy": mock_add,
            "selinux.port_modify_policy": mock_modify,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": True}):
            ret.update({"comment": "", "result": None})
            assert selinux.port_policy_present(name, "http_cache_port_t") == ret

        with patch.dict(selinux.__opts__, {"test": False}):
            ret.update(
                {
                    "comment": "",
                    "changes": {
                        "old": None,
                        "new": {
                            "sel_type": "http_cache_port_t",
                            "protocol": "tcp",
                            "port": "8080",
                        },
                    },
                    "result": True,
                }
            )
            assert selinux.port_policy_present(name, "http_cache_port_t") == ret

    # Test modifying policy to a new sel_type
    mock_add = MagicMock(return_value={"retcode": 0})
    mock_modify = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(
        side_effect=[
            None,
            None,
            {"sel_type": "http_cache_port_t", "protocol": "tcp", "port": "8080"},
            {"sel_type": "http_port_t", "protocol": "tcp", "port": "8080"},
        ]
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_add_policy": mock_add,
            "selinux.port_modify_policy": mock_modify,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": True}):
            ret.update({"comment": "", "changes": {}, "result": None})
            assert selinux.port_policy_present(name, "http_port_t") == ret

        with patch.dict(selinux.__opts__, {"test": False}):
            ret.update(
                {
                    "comment": "",
                    "changes": {
                        "old": {
                            "sel_type": "http_cache_port_t",
                            "protocol": "tcp",
                            "port": "8080",
                        },
                        "new": {
                            "sel_type": "http_port_t",
                            "protocol": "tcp",
                            "port": "8080",
                        },
                    },
                    "result": True,
                }
            )
            assert selinux.port_policy_present(name, "http_port_t") == ret

    # Test adding new port policy with custom name and using protocol and port parameters
    mock_add = MagicMock(return_value={"retcode": 0})
    mock_modify = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(
        side_effect=[
            None,
            None,
            {"sel_type": "http_cache_port_t", "protocol": "tcp", "port": "8081"},
        ]
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_add_policy": mock_add,
            "selinux.port_modify_policy": mock_modify,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": False}):
            ret.update(
                {
                    "name": "required_protocol_port",
                    "comment": "",
                    "changes": {
                        "old": None,
                        "new": {
                            "sel_type": "http_cache_port_t",
                            "protocol": "tcp",
                            "port": "8081",
                        },
                    },
                    "result": True,
                }
            )
            assert (
                selinux.port_policy_present(
                    "required_protocol_port",
                    "http_cache_port_t",
                    protocol="tcp",
                    port="8081",
                )
                == ret
            )

    # Test failure of adding new policy
    mock_add = MagicMock(return_value={"retcode": 1})
    mock_modify = MagicMock(return_value={"retcode": 1})
    mock_get = MagicMock(return_value=None)
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_add_policy": mock_add,
            "selinux.port_modify_policy": mock_modify,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": False}):
            comt = "Error adding new policy: {'retcode': 1}"
            ret.update({"name": name, "comment": comt, "changes": {}, "result": False})
            assert selinux.port_policy_present(name, "http_cache_port_t") == ret


def test_port_policy_absent():
    """
    Test to delete an SELinux port.
    """
    name = "tcp/8080"
    protocol = "tcp"
    port = "8080"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Test policy already removed
    mock_delete = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(return_value=None)
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_delete_policy": mock_delete,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": False}):
            comt = (
                f'SELinux policy for "{name}" already absent '
                + 'with specified sel_type "http_cache_port_t", protocol "None" '
                + 'and port "None".'
            )
            ret.update({"comment": comt, "changes": {}, "result": True})
            assert selinux.port_policy_absent(name, "http_cache_port_t") == ret

            comt = (
                'SELinux policy for "name" already absent '
                + f'with specified sel_type "http_cache_port_t", protocol "{protocol}" '
                + f'and port "{port}".'
            )
            ret.update({"comment": comt, "changes": {}, "result": True, "name": "name"})
            assert (
                selinux.port_policy_absent("name", "http_cache_port_t", protocol, port)
                == ret
            )
            ret.update({"name": name})

    # Test removing a policy
    mock_delete = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(
        side_effect=[
            {"sel_type": "http_cache_port_t", "protocol": "tcp", "port": "8080"},
            {"sel_type": "http_cache_port_t", "protocol": "tcp", "port": "8080"},
            None,
        ]
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_delete_policy": mock_delete,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": True}):
            ret.update({"comment": "", "result": None})
            assert selinux.port_policy_absent(name, "http_cache_port_t") == ret

        with patch.dict(selinux.__opts__, {"test": False}):
            ret.update(
                {
                    "comment": "",
                    "changes": {
                        "old": {
                            "sel_type": "http_cache_port_t",
                            "protocol": "tcp",
                            "port": "8080",
                        },
                        "new": None,
                    },
                    "result": True,
                }
            )
            assert selinux.port_policy_absent(name, "http_cache_port_t") == ret

    # Test removing a policy using custom name and with protocol and port parameters
    mock_delete = MagicMock(return_value={"retcode": 0})
    mock_get = MagicMock(
        side_effect=[
            {"sel_type": "http_cache_port_t", "protocol": "tcp", "port": "8081"},
            None,
        ]
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_delete_policy": mock_delete,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": False}):
            ret.update(
                {
                    "name": "required_protocol_port",
                    "comment": "",
                    "changes": {
                        "old": {
                            "sel_type": "http_cache_port_t",
                            "protocol": "tcp",
                            "port": "8081",
                        },
                        "new": None,
                    },
                    "result": True,
                }
            )
            assert (
                selinux.port_policy_absent(
                    "required_protocol_port",
                    "http_cache_port_t",
                    protocol="tcp",
                    port="8081",
                )
                == ret
            )

    # Test failure to delete a policy
    mock_delete = MagicMock(return_value={"retcode": 2})
    mock_get = MagicMock(
        return_value={
            "sel_type": "http_cache_port_t",
            "protocol": "tcp",
            "port": "8080",
        }
    )
    with patch.dict(
        selinux.__salt__,
        {
            "selinux.port_get_policy": mock_get,
            "selinux.port_delete_policy": mock_delete,
        },
    ):
        with patch.dict(selinux.__opts__, {"test": False}):
            comt = "Error deleting policy: {'retcode': 2}"
            ret.update({"name": name, "comment": comt, "changes": {}, "result": False})
            assert selinux.port_policy_absent(name, "http_cache_port_t") == ret



"""
Grains include tests
"""

import os


def test_grains_package(salt_call_cli):
    """
    An integration test for the package grain
    is a bit overkill, but it is necessary currently
    because the onedir package is only tested with
    integration tests, so I want to ensure its working
    correctly. Once the onedir package is tested with unit
    and functional tests this test can be removed since
    there is plenty of test coverage in both unit and functional
    for this new grain.
    """
    ret = salt_call_cli.run("grains.get", "package")
    assert ret.returncode == 0
    assert ret.data
    if os.environ.get("ONEDIR_TESTRUN", "0") == "0":
        assert ret.data == "pip"
    else:
        assert ret.data == "onedir"



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.sysctl as sysctl
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {sysctl: {}}


def test_empty_config_file_and_value_not_found():
    """
    Test sysctl.present for an unknown sysctl, not present in config file
    """
    name = "some.unknown.oid"
    value = "1"
    comment = f"Sysctl option {name} would be changed to {value}"

    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    with patch.dict(sysctl.__opts__, {"test": True}):
        mock_show = MagicMock(return_value={})
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock_show}):
            mock_get = MagicMock(return_value="")
            with patch.dict(sysctl.__salt__, {"sysctl.get": mock_get}):
                assert sysctl.present(name, value) == ret


def test_inaccessible_config_file():
    """
    Test sysctl.present with a config file that cannot be opened
    """
    name = "some.unknown.oid"
    value = "1"
    config = "/etc/sysctl.conf"
    comment = (
        "Sysctl option {} might be changed, we failed to check "
        "config file at {}. The file is either unreadable, or "
        "missing.".format(name, config)
    )
    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    with patch.dict(sysctl.__opts__, {"test": True}):
        mock = MagicMock(return_value=None)
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock}):
            assert sysctl.present(name, value) == ret


def test_to_be_changed_not_configured():
    """
    Test sysctl.present for a sysctl that isn't in the config file and must
    be changed.
    """
    name = "vfs.usermount"
    value = "1"
    comment = f"Sysctl option {name} set to be changed to {value}"
    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    def mock_current(config_file=None):
        """
        Mock return value for __salt__.
        """
        if config_file is None:
            return {name: "0"}
        return {}

    with patch.dict(sysctl.__opts__, {"test": True}):
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock_current}):
            mock_get = MagicMock(return_value="0")
            with patch.dict(sysctl.__salt__, {"sysctl.get": mock_get}):
                assert sysctl.present(name, value) == ret


def test_not_to_be_changed_not_configured():
    """
    Test sysctl.present for a sysctl that isn't in the config file but
    already has the correct value
    """
    name = "some.unknown.oid"
    value = "1"
    comment = (
        "Sysctl value is currently set on the running system but "
        "not in a config file. Sysctl option {} set to be "
        "changed to 1 in config file.".format(name)
    )

    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    def mock_current(config_file=None):
        if config_file is None:
            return {name: value}
        return {}

    with patch.dict(sysctl.__opts__, {"test": True}):
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock_current}):
            mock_get = MagicMock(return_value=value)
            with patch.dict(sysctl.__salt__, {"sysctl.get": mock_get}):
                assert sysctl.present(name, value) == ret


def test_configured_but_unknown():
    """
    Test sysctl.present for a sysctl that is already configured but is
    not known by the system.  For example, a sysctl used by a kernel module
    that isn't loaded.
    """
    name = "vfs.usermount"
    value = "1"
    comment = (
        "Sysctl value {0} is present in configuration file but is not "
        "present in the running config. The value {0} is set to be "
        "changed to {1}".format(name, value)
    )
    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    def mock_config(config_file=None):
        if config_file is None:
            return {}
        return {name: value}

    with patch.dict(sysctl.__opts__, {"test": True}):
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock_config}):
            mock_get = MagicMock(return_value="")
            with patch.dict(sysctl.__salt__, {"sysctl.get": mock_get}):
                assert sysctl.present(name, value) == ret


def test_no_change():
    """
    Test sysctl.present for an already-configured value
    """
    name = "vfs.usermount"
    value = "1"
    comment = f"Sysctl value {name} = {value} is already set"
    ret = {"name": name, "result": True, "changes": {}, "comment": comment}

    def mock_config(config_file=None):
        if config_file is None:
            return {}
        return {name: value}

    with patch.dict(sysctl.__opts__, {"test": True}):
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock_config}):
            mock_get = MagicMock(return_value=value)
            with patch.dict(sysctl.__salt__, {"sysctl.get": mock_get}):
                assert sysctl.present(name, value) == ret


def test_change():
    """
    Test sysctl.present for a value whose configuration must change
    """
    name = "vfs.usermount"
    old_value = "2"
    value = "1"
    comment = f"Sysctl option {name} would be changed to {value}"
    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    def mock_config(config_file=None):
        if config_file is None:
            return {name: old_value}
        return {name: old_value}

    with patch.dict(sysctl.__opts__, {"test": True}):
        with patch.dict(sysctl.__salt__, {"sysctl.show": mock_config}):
            mock_get = MagicMock(return_value=old_value)
            with patch.dict(sysctl.__salt__, {"sysctl.get": mock_get}):
                assert sysctl.present(name, value) == ret


def test_failed_to_set():
    """
    Test sysctl.present when the sysctl command fails to change a value
    """
    name = "net.isr.maxthreads"
    value = "8"
    comment = f"Failed to set {name} to {value}: "
    ret = {"name": name, "result": False, "changes": {}, "comment": comment}

    with patch.dict(sysctl.__opts__, {"test": False}):
        mock = MagicMock(side_effect=CommandExecutionError)
        with patch.dict(sysctl.__salt__, {"sysctl.persist": mock}):
            assert sysctl.present(name, value) == ret


def test_already_set():
    """
    Test sysctl.present when the value is already set
    """
    name = "vfs.usermount"
    value = "1"
    comment = f"Sysctl value {name} = {value} is already set"
    ret = {"name": name, "result": True, "changes": {}, "comment": comment}
    with patch.dict(sysctl.__opts__, {"test": False}):
        mock = MagicMock(return_value="Already set")
        with patch.dict(sysctl.__salt__, {"sysctl.persist": mock}):
            assert sysctl.present(name, value) == ret


def test_updated():
    """
    Test sysctl.present when the value is not already set
    """
    name = "vfs.usermount"
    value = "1"
    comment = f"Updated sysctl value {name} = {value}"
    changes = {name: value}
    ret = {"name": name, "result": True, "changes": changes, "comment": comment}
    with patch.dict(sysctl.__opts__, {"test": False}):
        mock = MagicMock(return_value="Updated")
        with patch.dict(sysctl.__salt__, {"sysctl.persist": mock}):
            assert sysctl.present(name, value) == ret



import pytest

import salt.returners.redis_return as redis_return
from tests.support.helpers import patch


@pytest.fixture
def configure_loader_modules():
    return {
        redis_return: {
            "__opts__": {
                "redis.host": "the mostest",
                "redis.port": 42,
                "redis.unix_socket_path": "the road less traveled",
                "redis.db": "13",
                "redis.password": "shibboleth",
                "redis.startup_nodes": {"lymph": "OK"},
                "redis.skip_full_coverage_check": True,
            }
        }
    }


@pytest.fixture(autouse=True)
def clean_REDIS_POOL():
    # Any unit test will want this value reset to None,
    # otherwise the values will remain set from other tests
    redis_return.REDIS_POOL = None  # pylint: disable=unmocked-patch


@pytest.fixture
def mock_strict_redis():
    with patch.object(redis_return, "redis", create=True):
        with patch(
            "salt.returners.redis_return.redis.StrictRedis"
        ) as fake_strict_redis:
            yield fake_strict_redis


@pytest.fixture
def proxy_platform():
    with patch("salt.utils.platform.is_proxy", autospec=True, return_value=True):
        yield


@pytest.fixture
def non_proxy_platform():
    with patch("salt.utils.platform.is_proxy", autospec=True, return_value=False):
        yield


@pytest.fixture
def mock_returner_options():
    options = {
        "host": "some other hosty mchostface",
        "port": 99,
        "unix_socket_path": "something socket",
        "db": "cooper",
        "password": "super secret!",
    }
    with patch(
        "salt.returners.get_returner_options", autospec=True, return_value=options
    ):
        yield options


def test_when_platform_is_proxy_redis_should_use_opts_values(
    proxy_platform, mock_strict_redis
):
    redis_return._get_serv()

    mock_strict_redis.assert_called_once_with(
        host="the mostest",
        port=42,
        unix_socket_path="the road less traveled",
        db="13",
        password="shibboleth",
        decode_responses=True,
    )


def test_when_platform_is_proxy_and_no_opts_are_set_fallback_values_should_be_used(
    proxy_platform, mock_strict_redis
):
    with patch.dict(redis_return.__opts__, {}, clear=True):
        redis_return._get_serv()

    mock_strict_redis.assert_called_once_with(
        host="salt",
        port=6379,
        unix_socket_path=None,
        db="0",
        password=None,
        decode_responses=True,
    )


def test_when_platform_is_not_proxy_it_should_use_returner_opts(
    non_proxy_platform, mock_strict_redis, mock_returner_options
):
    redis_return._get_serv()

    mock_strict_redis.assert_called_once_with(
        host="some other hosty mchostface",
        port=99,
        unix_socket_path="something socket",
        db="cooper",
        password="super secret!",
        decode_responses=True,
    )



import shutil

import pytest
from saltfactories.utils import random_string

from tests.conftest import FIPS_TESTRUN


@pytest.fixture(scope="module")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="module")
def peer_salt_master_config(pillar_state_tree):
    return {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "peer": {
            ".*": ["test.ping"],
            "peer-comm-minion.*": [
                {
                    "G@hello_peer:beer": ["grains.get"],
                }
            ],
        },
    }


@pytest.fixture(scope="module")
def peer_salt_master(
    salt_factories, pillar_state_tree, vault_port, peer_salt_master_config
):
    factory = salt_factories.salt_master_daemon(
        random_string("peer-comm-master", uppercase=False),
        defaults=peer_salt_master_config,
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "publish_signing_algorithm": (
                "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
            ),
        },
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="module")
def peer_salt_minion_1(peer_salt_master):
    assert peer_salt_master.is_running()
    factory = peer_salt_master.salt_minion_daemon(
        random_string("peer-comm-minion-1", uppercase=False),
        defaults={"open_mode": True, "grains": {"hello_peer": "beer"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="module")
def peer_salt_minion_2(peer_salt_master):
    assert peer_salt_master.is_running()
    factory = peer_salt_master.salt_minion_daemon(
        random_string("peer-comm-minion-2", uppercase=False),
        defaults={"open_mode": True},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="module")
def peer_salt_minion_3(peer_salt_master):
    assert peer_salt_master.is_running()
    factory = peer_salt_master.salt_minion_daemon(
        random_string("peer-comm-minion-3", uppercase=False),
        defaults={"open_mode": True},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.mark.parametrize(
    "source,target", ((x, y) for x in range(1, 4) for y in range(1, 4) if x != y)
)
def test_peer_communication(source, target, request, grains):
    if grains["os"] == "Fedora" and grains["osmajorrelease"] >= 40:
        pytest.skip(f"Temporary skip on {grains['osfinger']}")
    cli = request.getfixturevalue(f"peer_salt_minion_{source}").salt_call_cli()
    tgt = request.getfixturevalue(f"peer_salt_minion_{target}").id
    ret = cli.run("publish.publish", tgt, "test.ping")
    assert ret.returncode == 0
    assert ret.data
    assert tgt in ret.data
    assert ret.data[tgt] is True


def test_peer_communication_denied(peer_salt_minion_1, peer_salt_minion_2):
    tgt = peer_salt_minion_2.id
    ret = peer_salt_minion_1.salt_call_cli().run(
        "publish.publish", tgt, "cmd.run", "echo pwned"
    )
    assert ret.returncode == 0
    assert ret.data == {}


@pytest.mark.parametrize("source", [2, 3])
def test_peer_communication_limited_target_allowed(source, peer_salt_minion_1, request):
    cli = request.getfixturevalue(f"peer_salt_minion_{source}").salt_call_cli()
    tgt = peer_salt_minion_1.id
    ret = cli.run("publish.publish", tgt, "grains.get", "hello_peer")
    assert ret.returncode == 0
    assert ret.data
    assert tgt in ret.data
    assert ret.data[tgt] == "beer"


@pytest.mark.parametrize(
    "source,target", ((x, y) for x in range(1, 4) for y in range(2, 4) if x != y)
)
def test_peer_communication_limited_target_denied(source, target, request):
    cli = request.getfixturevalue(f"peer_salt_minion_{source}").salt_call_cli()
    tgt = request.getfixturevalue(f"peer_salt_minion_{target}").id
    ret = cli.run("publish.publish", tgt, "grains.get", "hello_peer")
    assert ret.returncode == 0
    assert ret.data == {}



"""
Support for BambooHR

.. versionadded:: 2015.8.0

Requires a ``subdomain`` and an ``apikey`` in ``/etc/salt/minion``:

.. code-block:: yaml

    bamboohr:
      apikey: 012345678901234567890
      subdomain: mycompany
"""

import logging
import xml.etree.ElementTree as ET

import salt.utils.http
import salt.utils.yaml

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if apache is installed
    """
    if _apikey():
        return True
    return (
        False,
        'The API key was not specified. Please specify it using the "apikey" config.',
    )


def _apikey():
    """
    Get the API key
    """
    return __opts__.get("bamboohr", {}).get("apikey", None)


def list_employees(order_by="id"):
    """
    Show all employees for this company.

    CLI Example:

    .. code-block:: bash

        salt myminion bamboohr.list_employees

    By default, the return data will be keyed by ID. However, it can be ordered
    by any other field. Keep in mind that if the field that is chosen contains
    duplicate values (i.e., location is used, for a company which only has one
    location), then each duplicate value will be overwritten by the previous.
    Therefore, it is advisable to only sort by fields that are guaranteed to be
    unique.

    CLI Examples:

    .. code-block:: bash

        salt myminion bamboohr.list_employees order_by=id
        salt myminion bamboohr.list_employees order_by=displayName
        salt myminion bamboohr.list_employees order_by=workEmail
    """
    ret = {}
    status, result = _query(action="employees", command="directory")
    root = ET.fromstring(result)
    for cat in root:
        if cat.tag != "employees":
            continue
        for item in cat:
            emp_id = next(iter(item.values()))
            emp_ret = {"id": emp_id}
            for details in item:
                emp_ret[next(iter(details.values()))] = details.text
            ret[emp_ret[order_by]] = emp_ret
    return ret


def show_employee(emp_id, fields=None):
    """
    Show all employees for this company.

    CLI Example:

    .. code-block:: bash

        salt myminion bamboohr.show_employee 1138

    By default, the fields normally returned from bamboohr.list_employees are
    returned. These fields are:

        - canUploadPhoto
        - department
        - displayName
        - firstName
        - id
        - jobTitle
        - lastName
        - location
        - mobilePhone
        - nickname
        - photoUploaded
        - photoUrl
        - workEmail
        - workPhone
        - workPhoneExtension

    If needed, a different set of fields may be specified, separated by commas:

    CLI Example:

    .. code-block:: bash

        salt myminion bamboohr.show_employee 1138 displayName,dateOfBirth

    A list of available fields can be found at
    http://www.bamboohr.com/api/documentation/employees.php
    """
    ret = {}
    if fields is None:
        fields = ",".join(
            (
                "canUploadPhoto",
                "department",
                "displayName",
                "firstName",
                "id",
                "jobTitle",
                "lastName",
                "location",
                "mobilePhone",
                "nickname",
                "photoUploaded",
                "photoUrl",
                "workEmail",
                "workPhone",
                "workPhoneExtension",
            )
        )

    status, result = _query(action="employees", command=emp_id, args={"fields": fields})

    root = ET.fromstring(result)

    ret = {"id": emp_id}
    for item in root:
        ret[next(iter(item.values()))] = item.text
    return ret


def update_employee(emp_id, key=None, value=None, items=None):
    """
    Update one or more items for this employee. Specifying an empty value will
    clear it for that employee.

    CLI Examples:

    .. code-block:: bash

        salt myminion bamboohr.update_employee 1138 nickname Curly
        salt myminion bamboohr.update_employee 1138 nickname ''
        salt myminion bamboohr.update_employee 1138 items='{"nickname": "Curly"}
        salt myminion bamboohr.update_employee 1138 items='{"nickname": ""}
    """
    if items is None:
        if key is None or value is None:
            return {"Error": "At least one key/value pair is required"}
        items = {key: value}
    elif isinstance(items, str):
        items = salt.utils.yaml.safe_load(items)

    xml_items = ""
    for pair in items:
        xml_items += f'<field id="{pair}">{items[pair]}</field>'
    xml_items = f"<employee>{xml_items}</employee>"

    status, result = _query(
        action="employees",
        command=emp_id,
        data=xml_items,
        method="POST",
    )

    return show_employee(emp_id, ",".join(items.keys()))


def list_users(order_by="id"):
    """
    Show all users for this company.

    CLI Example:

    .. code-block:: bash

        salt myminion bamboohr.list_users

    By default, the return data will be keyed by ID. However, it can be ordered
    by any other field. Keep in mind that if the field that is chosen contains
    duplicate values (i.e., location is used, for a company which only has one
    location), then each duplicate value will be overwritten by the previous.
    Therefore, it is advisable to only sort by fields that are guaranteed to be
    unique.

    CLI Examples:

    .. code-block:: bash

        salt myminion bamboohr.list_users order_by=id
        salt myminion bamboohr.list_users order_by=email
    """
    ret = {}
    status, result = _query(action="meta", command="users")
    root = ET.fromstring(result)
    for user in root:
        user_id = None
        user_ret = {}
        for item in user.items():
            user_ret[item[0]] = item[1]
            if item[0] == "id":
                user_id = item[1]
        for item in user:
            user_ret[item.tag] = item.text
        ret[user_ret[order_by]] = user_ret
    return ret


def list_meta_fields():
    """
    Show all meta data fields for this company.

    CLI Example:

    .. code-block:: bash

        salt myminion bamboohr.list_meta_fields
    """
    ret = {}
    status, result = _query(action="meta", command="fields")
    root = ET.fromstring(result)
    for field in root:
        field_id = None
        field_ret = {"name": field.text}
        for item in field.items():
            field_ret[item[0]] = item[1]
            if item[0] == "id":
                field_id = item[1]
        ret[field_id] = field_ret
    return ret


def _query(action=None, command=None, args=None, method="GET", data=None):
    """
    Make a web call to BambooHR

    The password can be any random text, so we chose Salty text.
    """
    subdomain = __opts__.get("bamboohr", {}).get("subdomain", None)
    path = f"https://api.bamboohr.com/api/gateway.php/{subdomain}/v1/"

    if action:
        path += action

    if command:
        path += f"/{command}"

    log.debug("BambooHR URL: %s", path)

    if not isinstance(args, dict):
        args = {}

    return_content = None
    result = salt.utils.http.query(
        path,
        method,
        username=_apikey(),
        password="saltypork",
        params=args,
        data=data,
        decode=False,
        text=True,
        status=True,
        opts=__opts__,
    )
    log.debug("BambooHR Response Status Code: %s", result["status"])

    return [result["status"], result["text"]]



"""
    :codeauthor: :email:`Shane Lee <slee@saltstack.com>`
"""

import pytest

import salt.grains.disks as disks
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {
        disks: {"__salt__": {}},
    }


def test__windows_disks_dict():
    """
    Test grains._windows_disks with a single disk returned as a dict
    Should return 1 disk and no ssds
    """
    devices = {"DeviceID": 0, "MediaType": "HDD"}
    mock_powershell = MagicMock(return_value=devices)

    with patch.dict(disks.__salt__, {"cmd.powershell": mock_powershell}):
        result = disks._windows_disks()
        expected = {"disks": ["\\\\.\\PhysicalDrive0"], "ssds": []}
        assert result == expected


def test__windows_disks_list():
    """
    test grains._windows_disks with multiple disks and types as a list of dicts
    Should return 4 disks and 1 ssd
    """
    devices = [
        {"DeviceID": 0, "MediaType": "SSD"},
        {"DeviceID": 1, "MediaType": "HDD"},
        {"DeviceID": 2, "MediaType": "HDD"},
        {"DeviceID": 3, "MediaType": "HDD"},
    ]
    mock_powershell = MagicMock(return_value=devices)

    with patch.dict(disks.__salt__, {"cmd.powershell": mock_powershell}):
        result = disks._windows_disks()
        expected = {
            "disks": [
                "\\\\.\\PhysicalDrive0",
                "\\\\.\\PhysicalDrive1",
                "\\\\.\\PhysicalDrive2",
                "\\\\.\\PhysicalDrive3",
            ],
            "ssds": ["\\\\.\\PhysicalDrive0"],
        }
        assert result == expected


def test__windows_disks_empty():
    """
    Test grains._windows_disks when nothing is returned
    Should return empty lists
    """
    devices = {}
    mock_powershell = MagicMock(return_value=devices)

    with patch.dict(disks.__salt__, {"cmd.powershell": mock_powershell}):
        expected = {"disks": [], "ssds": []}
        result = disks._windows_disks()
        assert result == expected


def test__linux_disks():
    """
    Test grains._linux_disks, normal return
    Should return a populated dictionary
    """

    files = [
        "/sys/block/asm!.asm_ctl_vbg0",
        "/sys/block/dm-0",
        "/sys/block/loop0",
        "/sys/block/ram0",
        "/sys/block/sda",
        "/sys/block/sdb",
        "/sys/block/vda",
    ]
    links = [
        "../devices/virtual/block/asm!.asm_ctl_vbg0",
        "../devices/virtual/block/dm-0",
        "../devices/virtual/block/loop0",
        "../devices/virtual/block/ram0",
        "../devices/pci0000:00/0000:00:1f.2/ata1/host0/target0:0:0/0:0:0:0/block/sda",
        "../devices/pci0000:35/0000:35:00.0/0000:36:00.0/host2/target2:1:0/2:1:0:0/block/sdb",
        "../devices/pci0000L00:0000:00:05.0/virtio2/block/vda",
    ]
    contents = [
        "1",
        "1",
        "1",
        "0",
        "1",
        "1",
        "1",
    ]

    patch_glob = patch("glob.glob", autospec=True, return_value=files)
    patch_readlink = patch("salt.utils.path.readlink", autospec=True, side_effect=links)
    patch_fopen = patch("salt.utils.files.fopen", mock_open(read_data=contents))
    with patch_glob, patch_readlink, patch_fopen:
        ret = disks._linux_disks()

    assert ret == {"disks": ["sda", "sdb", "vda"], "ssds": []}, ret



"""
    Test cases for salt.utils.etcd_util

    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.utils.etcd_util as etcd_util
from tests.support.mock import MagicMock, patch

if etcd_util.HAS_ETCD_V2:
    import etcd
    from urllib3.exceptions import MaxRetryError, ReadTimeoutError


def _version_id(value):
    return "etcd-v2" if value else "etcd-v3"


@pytest.fixture(scope="module", params=(True, False), ids=_version_id)
def use_v2(request):
    if request.param and not etcd_util.HAS_ETCD_V2:
        pytest.skip("No etcd library installed")
    if not request.param and not etcd_util.HAS_ETCD_V3:
        pytest.skip("No etcd3 library installed")
    return request.param


@pytest.fixture(scope="module")
def client_name(use_v2):
    if use_v2:
        return "etcd.Client"
    return "etcd3.Client"


def test_read(client_name, use_v2):
    """
    Test to make sure we interact with etcd correctly
    """
    with patch(client_name, autospec=True) as mock:
        etcd_client = mock.return_value
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )
        if use_v2:
            etcd_return = MagicMock(value="salt")
            etcd_client.read.return_value = etcd_return

            assert client.read("/salt") == etcd_return
            etcd_client.read.assert_called_with(
                "/salt", recursive=False, wait=False, timeout=None
            )

            client.read("salt", True, True, 10, 5)
            etcd_client.read.assert_called_with(
                "salt", recursive=True, wait=True, timeout=10, waitIndex=5
            )

            etcd_client.read.side_effect = etcd.EtcdKeyNotFound
            with pytest.raises(etcd.EtcdKeyNotFound):
                client.read("salt")

            etcd_client.read.side_effect = etcd.EtcdConnectionFailed
            with pytest.raises(etcd.EtcdConnectionFailed):
                client.read("salt")

            etcd_client.read.side_effect = etcd.EtcdValueError
            with pytest.raises(etcd.EtcdValueError):
                client.read("salt")

            etcd_client.read.side_effect = ValueError
            with pytest.raises(ValueError):
                client.read("salt")

            etcd_client.read.side_effect = ReadTimeoutError(None, None, None)
            with pytest.raises(etcd.EtcdConnectionFailed):
                client.read("salt")

            etcd_client.read.side_effect = MaxRetryError(None, None)
            with pytest.raises(etcd.EtcdConnectionFailed):
                client.read("salt")
        else:
            etcd_return = MagicMock(kvs=[MagicMock(value="salt")])
            etcd_client.range.return_value = etcd_return
            assert client.read("/salt") == etcd_return.kvs
            etcd_client.range.assert_called_with("/salt", prefix=False)

            etcd_client.range.side_effect = Exception
            assert client.read("/salt") is None

            watcher_mock = MagicMock()
            with patch.object(etcd_client, "Watcher", return_value=watcher_mock):
                client.read("salt", True, True, 10, 5)
                etcd_client.range.assert_called_with("/salt", prefix=False)
                watcher_mock.watch_once.assert_called_with(timeout=10)

                watcher_mock.watch_once.side_effect = Exception
                assert client.read("salt", True, True, 10, 5) is None


def test_get(client_name, use_v2):
    """
    Test if it get a value from etcd, by direct path
    """
    with patch(client_name, autospec=True):
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        if use_v2:
            with patch.object(client, "read", autospec=True) as mock:
                mock.return_value = MagicMock(value="stack")
                assert client.get("salt") == "stack"
                mock.assert_called_with("salt")

                # iter(list(Exception)) works correctly with both mock<1.1 and mock>=1.1
                mock.side_effect = iter([etcd.EtcdKeyNotFound()])
                assert client.get("not-found") is None

                mock.side_effect = iter([etcd.EtcdConnectionFailed()])
                assert client.get("watching") is None

                # python 2.6 test
                mock.side_effect = ValueError
                assert client.get("not-found") is None

                mock.side_effect = Exception
                with pytest.raises(Exception):
                    client.get("some-error")
        else:
            with patch.object(client, "read", autospec=True) as mock:
                mock.return_value = [MagicMock(value="stack")]
                assert client.get("salt") == "stack"
                mock.assert_called_with("salt")

        # Get with recurse now delegates to client.tree
        with patch.object(client, "tree", autospec=True) as tree_mock:
            tree_mock.return_value = {"salt": "stack"}
            assert client.get("salt", recurse=True) == {"salt": "stack"}
            tree_mock.assert_called_with("salt")


def test_tree(use_v2, client_name):
    """
    Test recursive gets
    """
    with patch(client_name, autospec=True):
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        if use_v2:
            with patch.object(client, "read", autospec=True) as mock:
                c1, c2 = MagicMock(), MagicMock()
                c1.__iter__.return_value = [
                    MagicMock(key="/x/a", value="1"),
                    MagicMock(key="/x/b", value="2"),
                    MagicMock(key="/x/c", dir=True),
                ]
                c2.__iter__.return_value = [MagicMock(key="/x/c/d", value="3")]
                mock.side_effect = iter(
                    [MagicMock(children=c1), MagicMock(children=c2)]
                )
                assert client.tree("/x") == {"a": "1", "b": "2", "c": {"d": "3"}}
                mock.assert_any_call("/x")
                mock.assert_any_call("/x/c")

                # iter(list(Exception)) works correctly with both mock<1.1 and mock>=1.1
                mock.side_effect = iter([etcd.EtcdKeyNotFound()])
                assert client.tree("not-found") is None

                mock.side_effect = ValueError
                assert client.tree("/x") is None

                mock.side_effect = Exception
                with pytest.raises(Exception):
                    client.tree("some-error")
        else:
            with patch.object(client, "read", autospec=True) as mock:
                mock.return_value = [
                    MagicMock(key="/x/a", value="1"),
                    MagicMock(key="/x/b", value="2"),
                    MagicMock(key="/x/c/d", value="3"),
                ]
                assert client.tree("/x") == {"a": "1", "b": "2", "c": {"d": "3"}}
                mock.assert_called_with("/x", recurse=True)


def test_ls(use_v2, client_name):
    with patch(client_name, autospec=True):
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        if use_v2:
            with patch.object(client, "read", autospec=True) as mock:
                c1 = MagicMock()
                c1.__iter__.return_value = [
                    MagicMock(key="/x/a", value="1"),
                    MagicMock(key="/x/b", value="2"),
                    MagicMock(key="/x/c", dir=True),
                ]
                mock.return_value = MagicMock(children=c1)
                assert client.ls("/x") == {
                    "/x": {"/x/a": "1", "/x/b": "2", "/x/c/": {}}
                }
                mock.assert_called_with("/x")

                # iter(list(Exception)) works correctly with both mock<1.1 and mock>=1.1
                mock.side_effect = iter([etcd.EtcdKeyNotFound()])
                assert client.ls("/not-found") == {}

                mock.side_effect = Exception
                with pytest.raises(Exception):
                    client.tree("some-error")
        else:
            with patch.object(client, "read", autospec=True) as mock:
                mock.return_value = [
                    MagicMock(key="/x/a", value="1"),
                    MagicMock(key="/x/b", value="2"),
                    MagicMock(key="/x/c", value={"d": "3"}),
                ]
                assert client.ls("/x") == {
                    "/x": {"/x/a": "1", "/x/b": "2", "/x/c/": {}}
                }
                mock.assert_called_with("/x", recurse=True)


def test_write(use_v2, client_name):
    with patch(client_name, autospec=True) as mock:
        etcd_client = mock.return_value
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        if use_v2:
            etcd_client.write.return_value = MagicMock(value="salt")
            assert client.write("/some-key", "salt") == "salt"
            etcd_client.write.assert_called_with(
                "/some-key", "salt", ttl=None, dir=False
            )

            assert client.write("/some-key", "salt", ttl=5) == "salt"
            etcd_client.write.assert_called_with("/some-key", "salt", ttl=5, dir=False)

            etcd_client.write.return_value = MagicMock(dir=True)
            assert client.write("/some-dir", "salt", ttl=0, directory=True)
            etcd_client.write.assert_called_with("/some-dir", None, ttl=0, dir=True)

            # Check when a file is attempted to be written to a read-only root
            etcd_client.write.side_effect = etcd.EtcdRootReadOnly()
            assert client.write("/", "some-val", directory=False) is None

            # Check when a directory is attempted to be written to a read-only root
            etcd_client.write.side_effect = etcd.EtcdRootReadOnly()
            assert client.write("/", None, directory=True) is None

            # Check when a file is attempted to be written when unable to connect to the service
            etcd_client.write.side_effect = MaxRetryError(None, None)
            assert client.write("/some-key", "some-val", directory=False) is None

            # Check when a directory is attempted to be written when unable to connect to the service
            etcd_client.write.side_effect = MaxRetryError(None, None)
            assert client.write("/some-dir", None, directory=True) is None

            # Check when a file is attempted to be written to a directory that already exists (name-collision)
            etcd_client.write.side_effect = etcd.EtcdNotFile()
            assert client.write("/some-dir", "some-val", directory=False) is None

            # Check when a directory is attempted to be written to a file that already exists (name-collision)
            etcd_client.write.side_effect = etcd.EtcdNotDir()
            assert client.write("/some-key", None, directory=True) is None

            # Check when a directory is attempted to be written to a directory that already exists (update-ttl)
            etcd_client.write.side_effect = etcd.EtcdNotFile()
            assert client.write("/some-dir", None, directory=True)

            etcd_client.write.side_effect = ValueError
            assert client.write("/some-key", "some-val") is None

            etcd_client.write.side_effect = Exception
            with pytest.raises(Exception):
                client.set("some-key", "some-val")
        else:
            with pytest.raises(etcd_util.Etcd3DirectoryException):
                client.write("key", None, directory=True)

            with patch.object(client, "get", autospec=True) as get_mock:
                get_mock.return_value = "stack"
                assert client.write("salt", "stack") == "stack"
                etcd_client.put.assert_called_with("salt", "stack")

                lease_mock = MagicMock(ID=1)
                with patch.object(etcd_client, "Lease", return_value=lease_mock):
                    assert client.write("salt", "stack", ttl=5) == "stack"
                    etcd_client.put.assert_called_with("salt", "stack", lease=1)


def test_flatten(use_v2, client_name):
    with patch(client_name, autospec=True) as mock:
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )
        some_data = {
            "/x/y/a": "1",
            "x": {"y": {"b": "2"}},
            "m/j/": "3",
            "z": "4",
            "d": {},
        }

        result_path = {
            "/test/x/y/a": "1",
            "/test/x/y/b": "2",
            "/test/m/j": "3",
            "/test/z": "4",
            "/test/d": {},
        }

        result_nopath = {
            "/x/y/a": "1",
            "/x/y/b": "2",
            "/m/j": "3",
            "/z": "4",
            "/d": {},
        }

        result_root = {
            "/x/y/a": "1",
            "/x/y/b": "2",
            "/m/j": "3",
            "/z": "4",
            "/d": {},
        }

        assert client._flatten(some_data, path="/test") == result_path
        assert client._flatten(some_data, path="/") == result_root
        assert client._flatten(some_data) == result_nopath


def test_update(use_v2, client_name):
    with patch(client_name, autospec=True) as mock:
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )
        some_data = {
            "/x/y/a": "1",
            "x": {"y": {"b": "3"}},
            "m/j/": "3",
            "z": "4",
            "d": {},
        }

        result = {
            "/test/x/y/a": "1",
            "/test/x/y/b": "2",
            "/test/m/j": "3",
            "/test/z": "4",
            "/test/d": True,
        }

        flatten_result = {
            "/test/x/y/a": "1",
            "/test/x/y/b": "2",
            "/test/m/j": "3",
            "/test/z": "4",
            "/test/d": {},
        }
        client._flatten = MagicMock(return_value=flatten_result)

        assert client.update("/some/key", path="/blah") is None

        with patch.object(client, "write", autospec=True) as write_mock:

            def write_return(key, val, ttl=None, directory=None):
                return result.get(key, None)

            write_mock.side_effect = write_return
            if not use_v2:
                result.pop("/test/d")
            assert client.update(some_data, path="/test") == result
            client._flatten.assert_called_with(some_data, "/test")
            assert write_mock.call_count == 5 if use_v2 else 4


def test_rm(use_v2, client_name):
    with patch(client_name, autospec=True) as mock:
        etcd_client = mock.return_value
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        if use_v2:
            etcd_client.delete.return_value = True
            assert client.rm("/some-key")
            etcd_client.delete.assert_called_with("/some-key", recursive=False)
            assert client.rm("/some-dir", recurse=True)
            etcd_client.delete.assert_called_with("/some-dir", recursive=True)

            etcd_client.delete.side_effect = etcd.EtcdNotFile()
            assert client.rm("/some-dir") is None

            etcd_client.delete.side_effect = etcd.EtcdDirNotEmpty()
            assert client.rm("/some-key") is None

            etcd_client.delete.side_effect = etcd.EtcdRootReadOnly()
            assert client.rm("/") is None

            etcd_client.delete.side_effect = ValueError
            assert client.rm("/some-dir") is None

            etcd_client.delete.side_effect = Exception
            with pytest.raises(Exception):
                client.rm("some-dir")
        else:
            etcd_client.delete_range.return_value = MagicMock(deleted=1)
            assert client.rm("/some-key")
            etcd_client.delete_range.assert_called_with("/some-key", prefix=False)

            etcd_client.delete_range.return_value = MagicMock(deleted=0)
            assert client.rm("/some-key", recurse=True) is None
            etcd_client.delete_range.assert_called_with("/some-key", prefix=True)

            delattr(etcd_client.delete_range.return_value, "deleted")
            assert not client.rm("/some-key")
            etcd_client.delete_range.assert_called_with("/some-key", prefix=False)


def test_watch(use_v2, client_name):
    with patch(client_name, autospec=True):
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        if use_v2:
            with patch.object(client, "read", autospec=True) as mock:
                mock.return_value = MagicMock(
                    value="stack", key="/some-key", modifiedIndex=1, dir=False
                )
                assert client.watch("/some-key") == {
                    "value": "stack",
                    "key": "/some-key",
                    "mIndex": 1,
                    "changed": True,
                    "dir": False,
                }
                mock.assert_called_with(
                    "/some-key",
                    wait=True,
                    recurse=False,
                    timeout=0,
                    start_revision=None,
                )

                mock.side_effect = iter(
                    [etcd_util.EtcdUtilWatchTimeout, mock.return_value]
                )
                assert client.watch("/some-key") == {
                    "value": "stack",
                    "changed": False,
                    "mIndex": 1,
                    "key": "/some-key",
                    "dir": False,
                }

                mock.side_effect = iter(
                    [etcd_util.EtcdUtilWatchTimeout, etcd.EtcdKeyNotFound]
                )
                assert client.watch("/some-key") == {
                    "value": None,
                    "changed": False,
                    "mIndex": 0,
                    "key": "/some-key",
                    "dir": False,
                }

                mock.side_effect = iter([etcd_util.EtcdUtilWatchTimeout, ValueError])
                assert client.watch("/some-key") == {}

                mock.side_effect = None
                mock.return_value = MagicMock(
                    value="stack", key="/some-key", modifiedIndex=1, dir=True
                )
                assert client.watch(
                    "/some-dir", recurse=True, timeout=5, start_revision=10
                ) == {
                    "value": "stack",
                    "key": "/some-key",
                    "mIndex": 1,
                    "changed": True,
                    "dir": True,
                }
                mock.assert_called_with(
                    "/some-dir", wait=True, recurse=True, timeout=5, start_revision=10
                )

                # iter(list(Exception)) works correctly with both mock<1.1 and mock>=1.1
                mock.side_effect = iter([MaxRetryError(None, None)])
                assert client.watch("/some-key") == {}

                mock.side_effect = iter([etcd.EtcdConnectionFailed()])
                assert client.watch("/some-key") is None

                mock.side_effect = None
                mock.return_value = None
                assert client.watch("/some-key") == {}
        else:
            with patch.object(client, "read", autospec=True) as mock:
                mock.return_value = MagicMock(
                    value="stack", key="/some-key", mod_revision=1
                )
                assert client.watch("/some-key") == {
                    "value": "stack",
                    "key": "/some-key",
                    "mIndex": 1,
                    "changed": True,
                    "dir": False,
                }
                mock.assert_called_with(
                    "/some-key",
                    wait=True,
                    recurse=False,
                    timeout=0,
                    start_revision=None,
                )
                mock.return_value = MagicMock(
                    value="stack", key="/some-key", mod_revision=1
                )
                assert client.watch(
                    "/some-key", recurse=True, timeout=5, start_revision=10
                ) == {
                    "value": "stack",
                    "key": "/some-key",
                    "mIndex": 1,
                    "changed": True,
                    "dir": False,
                }
                mock.assert_called_with(
                    "/some-key", wait=True, recurse=True, timeout=5, start_revision=10
                )

                mock.side_effect = None
                mock.return_value = None
                assert client.watch("/some-key") is None


def test_expand(use_v2, client_name):
    if use_v2:
        pytest.skip("No expand in etcd v2 adapter")

    with patch(client_name, autospec=True) as mock:
        client = etcd_util.get_conn(
            {"etcd.require_v2": use_v2, "etcd.encode_values": False}
        )

        some_data = {
            "/test/x/y/a": "1",
            "/test/x/y/b": "2",
            "/test/m/j": "3",
            "/test/z": "4",
        }

        result = {
            "test": {
                "x": {"y": {"a": "1", "b": "2"}},
                "m": {"j": "3"},
                "z": "4",
            },
        }

        assert client._expand(some_data) == result



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)

    tests.unit.config.schemas.test_ssh
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest

import salt.utils.stringutils
from salt.config.schemas import ssh as ssh_schemas
from salt.config.schemas.minion import MinionConfiguration
from salt.utils.versions import Version

try:
    import jsonschema
    import jsonschema.exceptions

    HAS_JSONSCHEMA = True
    JSONSCHEMA_VERSION = Version(jsonschema.__version__)
except ImportError:
    HAS_JSONSCHEMA = False
    JSONSCHEMA_VERSION = Version("0")


def test_config():
    config = ssh_schemas.RosterEntryConfig()

    expected = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": "Roster Entry",
        "description": "Salt SSH roster entry definition",
        "type": "object",
        "properties": {
            "host": {
                "title": "Host",
                "description": "The IP address or DNS name of the remote host",
                "type": "string",
                "pattern": r"^((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})|([A-Za-z0-9][A-Za-z0-9\.\-]{1,255}))$",
                "minLength": 1,
            },
            "port": {
                "description": "The target system's ssh port number",
                "title": "Port",
                "default": 22,
                "maximum": 65535,
                "minimum": 0,
                "type": "integer",
            },
            "user": {
                "default": "root",
                "type": "string",
                "description": "The user to log in as. Defaults to root",
                "title": "User",
                "minLength": 1,
            },
            "passwd": {
                "title": "Password",
                "type": "string",
                "description": "The password to log in with",
                "format": "secret",
                "minLength": 1,
            },
            "priv": {
                "type": "string",
                "description": (
                    "File path to ssh private key, defaults to salt-ssh.rsa"
                ),
                "title": "Private Key",
                "minLength": 1,
            },
            "priv_passwd": {
                "type": "string",
                "description": "Passphrase for private key file",
                "title": "Private Key passphrase",
                "format": "secret",
                "minLength": 1,
            },
            "sudo": {
                "default": False,
                "type": "boolean",
                "description": "run command via sudo. Defaults to False",
                "title": "Sudo",
            },
            "timeout": {
                "type": "integer",
                "description": (
                    "Number of seconds to wait for response when establishing an"
                    " SSH connection"
                ),
                "title": "Timeout",
            },
            "thin_dir": {
                "type": "string",
                "description": (
                    "The target system's storage directory for Salt components."
                    " Defaults to /tmp/salt-<hash>."
                ),
                "title": "Thin Directory",
            },
            # The actuall representation of the minion options would make this HUGE!
            "minion_opts": ssh_schemas.DictItem(
                title="Minion Options",
                description="Dictionary of minion options",
                properties=MinionConfiguration(),
            ).serialize(),
        },
        "anyOf": [{"required": ["passwd"]}, {"required": ["priv"]}],
        "required": ["host", "user"],
        "x-ordering": [
            "host",
            "port",
            "user",
            "passwd",
            "priv",
            "priv_passwd",
            "sudo",
            "timeout",
            "thin_dir",
            "minion_opts",
        ],
        "additionalProperties": False,
    }
    try:
        # Workaround for assertDictContainsSubset
        assert (
            dict(expected["properties"], **config.serialize()["properties"])
            == expected["properties"]
        )
        assert dict(expected, **config.serialize()) == expected
    except AssertionError:
        import salt.utils.json

        print(salt.utils.json.dumps(config.serialize(), indent=4))
        raise


@pytest.mark.skipif(
    HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
)
def test_config_validate():
    try:
        jsonschema.validate(
            {"host": "localhost", "user": "root", "passwd": "foo"},
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    try:
        jsonschema.validate(
            {"host": "127.0.0.1", "user": "root", "passwd": "foo"},
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    try:
        jsonschema.validate(
            {"host": "127.1.0.1", "user": "root", "priv": "foo", "passwd": "foo"},
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    try:
        jsonschema.validate(
            {"host": "127.1.0.1", "user": "root", "passwd": "foo", "sudo": False},
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    try:
        jsonschema.validate(
            {
                "host": "127.1.0.1",
                "user": "root",
                "priv": "foo",
                "passwd": "foo",
                "thin_dir": "/foo/bar",
            },
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    try:
        jsonschema.validate(
            {
                "host": "127.1.0.1",
                "user": "root",
                "passwd": "foo",
                "minion_opts": {"interface": "0.0.0.0"},
            },
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    with pytest.raises(jsonschema.exceptions.ValidationError) as excinfo:
        jsonschema.validate(
            {"host": "127.1.0.1", "user": "", "passwd": "foo"},
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    assert "is too short" in excinfo.value.message

    with pytest.raises(jsonschema.exceptions.ValidationError) as excinfo:
        jsonschema.validate(
            {
                "host": "127.1.0.1",
                "user": "root",
                "passwd": "foo",
                "minion_opts": {"interface": 0},
            },
            ssh_schemas.RosterEntryConfig.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    assert "is not of type" in excinfo.value.message


def test_roster_config():
    try:
        subset = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Roster Configuration",
            "description": "Roster entries definition",
            "type": "object",
            "patternProperties": {
                r"^([^:]+)$": ssh_schemas.RosterEntryConfig.serialize()
            },
            "additionalProperties": False,
        }
        superset = ssh_schemas.RosterItem.serialize()

        assert dict(superset, **subset) == superset
    except AssertionError:
        import salt.utils.json

        print(salt.utils.json.dumps(ssh_schemas.RosterItem.serialize(), indent=4))
        raise


@pytest.mark.skipif(
    HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
)
def test_roster_config_validate():
    try:
        jsonschema.validate(
            {"target-1": {"host": "localhost", "user": "root", "passwd": "foo"}},
            ssh_schemas.RosterItem.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    except jsonschema.exceptions.ValidationError as exc:
        pytest.fail(f"ValidationError raised: {exc}")

    with pytest.raises(jsonschema.exceptions.ValidationError) as excinfo:
        jsonschema.validate(
            {
                salt.utils.stringutils.to_str("target-1:1"): {
                    "host": "localhost",
                    "user": "root",
                    "passwd": "foo",
                }
            },
            ssh_schemas.RosterItem.serialize(),
            format_checker=jsonschema.FormatChecker(),
        )
    if JSONSCHEMA_VERSION < Version("2.6.0"):
        assert (
            "Additional properties are not allowed ('target-1:1' was unexpected)"
            in excinfo.value.message
        )
    else:
        assert "'target-1:1' does not match any of the regexes" in excinfo.value.message



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""

import pytest

import salt.grains.napalm as napalm_grains
import salt.proxy.napalm as napalm_proxy
import tests.support.napalm as napalm_test_support
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    test_device_cache = {
        "DRIVER": napalm_test_support.MockNapalmDevice(),
        "DRIVER_NAME": "cisco",
        "OS_VERSION": "1.2.3",
        "HOSTNAME": "test-device.com",
        "USERNAME": "admin",
    }

    test_cache = {"result": True, "out": napalm_test_support.TEST_FACTS}
    module_globals = {
        "__salt__": {
            "config.option": MagicMock(
                return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
            ),
            "file.file_exists": napalm_test_support.true,
            "file.join": napalm_test_support.join,
            "file.get_managed": napalm_test_support.get_managed_file,
            "random.hash": napalm_test_support.random_hash,
        }
    }
    device_cache_patcher = patch("salt.grains.napalm.DEVICE_CACHE", test_device_cache)
    grains_cache_patcher = patch("salt.grains.napalm.GRAINS_CACHE", test_cache)
    is_proxy_patcher = patch.object(
        napalm_grains.salt.utils.napalm, "is_proxy", MagicMock(return_value=True)
    )

    with device_cache_patcher, grains_cache_patcher, is_proxy_patcher:
        yield {napalm_grains: module_globals}


def test_os():
    ret = napalm_grains.getos(proxy=napalm_proxy)
    assert ret["os"] == "cisco"


def test_os_version():
    ret = napalm_grains.version(proxy=napalm_proxy)
    assert ret["version"] == "1.2.3"


def test_model():
    ret = napalm_grains.model(proxy=napalm_proxy)
    assert ret["model"] == "test_model"


def test_serial():
    ret = napalm_grains.serial(proxy=napalm_proxy)
    assert ret["serial"] == "123456"


def test_vendor():
    ret = napalm_grains.vendor(proxy=napalm_proxy)
    assert ret["vendor"] == "cisco"


def test_uptime():
    ret = napalm_grains.uptime(proxy=napalm_proxy)
    assert ret["uptime"] == "Forever"


def test_interfaces():
    ret = napalm_grains.interfaces(proxy=napalm_proxy)
    assert ret["interfaces"] == napalm_test_support.TEST_INTERFACES


def test_username():
    ret = napalm_grains.username(proxy=napalm_proxy)
    assert ret["username"] == "admin"


def test_hostname():
    ret = napalm_grains.hostname(proxy=napalm_proxy)
    assert ret["hostname"] == "test-device.com"


def test_host():
    ret = napalm_grains.host(proxy=napalm_proxy)
    assert ret["host"] == "test-device.com"



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    salt.config.schemas.ssh
    ~~~~~~~~~~~~~~~~~~~~~~~

    Salt SSH related configuration schemas
"""

from salt.config.schemas.minion import MinionConfiguration
from salt.utils.schema import (
    AnyOfItem,
    BooleanItem,
    DictItem,
    IntegerItem,
    PortItem,
    RequirementsItem,
    Schema,
    SecretItem,
    StringItem,
)


class RosterEntryConfig(Schema):
    """
    Schema definition of a Salt SSH Roster entry
    """

    title = "Roster Entry"
    description = "Salt SSH roster entry definition"

    host = StringItem(
        title="Host",
        description="The IP address or DNS name of the remote host",
        # Pretty naive pattern matching
        pattern=r"^((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})|([A-Za-z0-9][A-Za-z0-9\.\-]{1,255}))$",
        min_length=1,
        required=True,
    )
    port = PortItem(
        title="Port", description="The target system's ssh port number", default=22
    )
    user = StringItem(
        title="User",
        description="The user to log in as. Defaults to root",
        default="root",
        min_length=1,
        required=True,
    )
    passwd = SecretItem(
        title="Password", description="The password to log in with", min_length=1
    )
    priv = StringItem(
        title="Private Key",
        description="File path to ssh private key, defaults to salt-ssh.rsa",
        min_length=1,
    )
    priv_passwd = SecretItem(
        title="Private Key passphrase",
        description="Passphrase for private key file",
        min_length=1,
    )
    passwd_or_priv_requirement = AnyOfItem(
        items=(
            RequirementsItem(requirements=["passwd"]),
            RequirementsItem(requirements=["priv"]),
        )
    )(flatten=True)
    sudo = BooleanItem(
        title="Sudo",
        description="run command via sudo. Defaults to False",
        default=False,
    )
    timeout = IntegerItem(
        title="Timeout",
        description=(
            "Number of seconds to wait for response when establishing an SSH connection"
        ),
    )
    thin_dir = StringItem(
        title="Thin Directory",
        description=(
            "The target system's storage directory for Salt "
            "components. Defaults to /tmp/salt-<hash>."
        ),
    )
    minion_opts = DictItem(
        title="Minion Options",
        description="Dictionary of minion options",
        properties=MinionConfiguration(),
    )


class RosterItem(Schema):
    title = "Roster Configuration"
    description = "Roster entries definition"

    roster_entries = DictItem(pattern_properties={r"^([^:]+)$": RosterEntryConfig()})(
        flatten=True
    )



import logging
from collections import namedtuple

import pytest

import salt.utils.args
from salt.exceptions import SaltInvocationError
from tests.support.mock import DEFAULT, patch

log = logging.getLogger(__name__)


def test_condition_input_string():
    """
    Test passing a jid on the command line
    """
    cmd = salt.utils.args.condition_input(["*", "foo.bar", 20141020201325675584], None)
    assert isinstance(cmd[2], str)


def test_clean_kwargs():
    assert salt.utils.args.clean_kwargs(foo="bar") == {"foo": "bar"}
    assert salt.utils.args.clean_kwargs(__pub_foo="bar") == {}
    assert salt.utils.args.clean_kwargs(__foo_bar="gwar") == {}
    assert salt.utils.args.clean_kwargs(foo_bar="gwar") == {"foo_bar": "gwar"}


def test_get_function_argspec():
    class DummyClass:
        def __init__(self, first):
            pass

        def __call__(self, first):
            pass

    dummy_class = DummyClass("foo")

    def dummy_func_nodefault(first, second, third):
        pass

    def dummy_func_default(first, second, third="fourth"):
        pass

    def dummy_func_varargs_keywords(*args, **kwargs):
        pass

    def dummy_func_annotations(first: int, second: str) -> bool:
        pass

    _ArgSpec = namedtuple("ArgSpec", "args varargs keywords defaults")

    # Callable class instance
    expected_argspec = _ArgSpec(
        args=["first"],
        varargs=None,
        keywords=None,
        defaults=None,
    )
    ret = salt.utils.args.get_function_argspec(dummy_class)
    assert ret == expected_argspec

    # Function with no varargs, no keywords, and no defaults
    expected_argspec = _ArgSpec(
        args=["first", "second", "third"],
        varargs=None,
        keywords=None,
        defaults=None,
    )
    ret = salt.utils.args.get_function_argspec(dummy_func_nodefault)
    assert ret == expected_argspec

    # Function with no varargs, no keywords, and one default
    expected_argspec = _ArgSpec(
        args=["first", "second", "third"],
        varargs=None,
        keywords=None,
        defaults=("fourth",),
    )
    ret = salt.utils.args.get_function_argspec(dummy_func_default)
    assert ret == expected_argspec

    # Function with annotations
    expected_argspec = _ArgSpec(
        args=["first", "second"],
        varargs=None,
        keywords=None,
        defaults=None,
    )
    ret = salt.utils.args.get_function_argspec(dummy_func_annotations)
    assert ret == expected_argspec

    # Function with both varargs and keywords
    expected_argspec = _ArgSpec(
        args=[],
        varargs="args",
        keywords="kwargs",
        defaults=None,
    )
    ret = salt.utils.args.get_function_argspec(dummy_func_varargs_keywords)
    assert ret == expected_argspec

    # Test that is_class_method=True is respected. Note that we don't
    # actually have a decorated function from rest_tornado here to test
    # this case, but we're testing for the behavior we expect, which is
    # that the first argument is popped off of the args.
    expected_argspec = _ArgSpec(
        args=["second", "third"],
        varargs=None,
        keywords=None,
        defaults=None,
    )
    ret = salt.utils.args.get_function_argspec(
        dummy_func_nodefault, is_class_method=True
    )
    assert ret == expected_argspec


def test_parse_kwarg():
    ret = salt.utils.args.parse_kwarg("foo=bar")
    assert ret == ("foo", "bar")

    ret = salt.utils.args.parse_kwarg("foobar")
    assert ret == (None, None)


def test_arg_lookup():
    def dummy_func(first, second, third, fourth="fifth"):
        pass

    expected_dict = {
        "args": ["first", "second", "third"],
        "kwargs": {"fourth": "fifth"},
    }
    ret = salt.utils.args.arg_lookup(dummy_func)
    assert ret == expected_dict


def test_format_call():
    with patch("salt.utils.args.arg_lookup") as arg_lookup:

        def dummy_func(first=None, second=None, third=None):
            pass

        arg_lookup.return_value = {
            "args": ["first", "second", "third"],
            "kwargs": {},
        }
        get_function_argspec = DEFAULT
        get_function_argspec.return_value = namedtuple(
            "ArgSpec", "args varargs keywords defaults"
        )(
            args=["first", "second", "third", "fourth"],
            varargs=None,
            keywords=None,
            defaults=("fifth",),
        )

        # Make sure we raise an error if we don't pass in the requisite number of arguments
        with pytest.raises(SaltInvocationError):
            salt.utils.args.format_call(dummy_func, {"1": 2})

        # Make sure we warn on invalid kwargs
        with pytest.raises(SaltInvocationError):
            salt.utils.args.format_call(
                dummy_func, {"first": 2, "seconds": 2, "third": 3}
            )

        ret = salt.utils.args.format_call(
            dummy_func,
            {"first": 2, "second": 2, "third": 3},
            expected_extra_kws=("first", "second", "third"),
        )
        assert ret == {"args": [], "kwargs": {}}


def test_format_call_simple_args():
    def foo(one, two=2, three=3):
        pass

    assert salt.utils.args.format_call(foo, dict(one=10, two=20, three=30)) == {
        "args": [10],
        "kwargs": dict(two=20, three=30),
    }
    assert salt.utils.args.format_call(foo, dict(one=10, two=20)) == {
        "args": [10],
        "kwargs": dict(two=20, three=3),
    }
    assert salt.utils.args.format_call(foo, dict(one=2)) == {
        "args": [2],
        "kwargs": dict(two=2, three=3),
    }


def test_format_call_mimic_typeerror_exceptions():
    def foo(one, two=2, three=3):
        pass

    def foo2(one, two, three=3):
        pass

    with pytest.raises(SaltInvocationError) as exc:
        salt.utils.args.format_call(foo, dict(two=3))
    assert "foo takes at least 1 argument (0 given)" in str(exc)

    with pytest.raises(SaltInvocationError) as exc:
        salt.utils.args.format_call(foo2, dict(one=1))
    assert "foo2 takes at least 2 arguments (1 given)" in str(exc)


def test_argspec_report():
    def _test_spec(arg1, arg2, kwarg1=None):
        pass

    test_functions = {"test_module.test_spec": _test_spec}
    ret = salt.utils.args.argspec_report(test_functions, "test_module.test_spec")
    assert ret == {
        "test_module.test_spec": {
            "kwargs": None,
            "args": ["arg1", "arg2", "kwarg1"],
            "defaults": (None,),
            "varargs": None,
        }
    }


def test_test_mode():
    assert salt.utils.args.test_mode(test=True)
    assert salt.utils.args.test_mode(Test=True)
    assert salt.utils.args.test_mode(tEsT=True)


def test_parse_function_no_args():
    fun, args, kwargs = salt.utils.args.parse_function("amod.afunc()")
    assert fun == "amod.afunc"
    assert args == []
    assert kwargs == {}


def test_parse_function_args_only():
    fun, args, kwargs = salt.utils.args.parse_function("amod.afunc(str1, str2)")
    assert fun == "amod.afunc"
    assert args == ["str1", "str2"]
    assert kwargs == {}


def test_parse_function_kwargs_only():
    fun, args, kwargs = salt.utils.args.parse_function("amod.afunc(kw1=val1, kw2=val2)")
    assert fun == "amod.afunc"
    assert args == []
    assert kwargs == {"kw1": "val1", "kw2": "val2"}


def test_parse_function_args_kwargs():
    fun, args, kwargs = salt.utils.args.parse_function(
        "amod.afunc(str1, str2, kw1=val1, kw2=val2)"
    )
    assert fun == "amod.afunc"
    assert args == ["str1", "str2"]
    assert kwargs == {"kw1": "val1", "kw2": "val2"}


def test_parse_function_malformed_no_name():
    fun, args, kwargs = salt.utils.args.parse_function(
        "(str1, str2, kw1=val1, kw2=val2)"
    )
    assert fun is None
    assert args is None
    assert kwargs is None


def test_parse_function_malformed_not_fun_def():
    fun, args, kwargs = salt.utils.args.parse_function("foo bar, some=text")
    assert fun is None
    assert args is None
    assert kwargs is None


def test_parse_function_wrong_bracket_style():
    fun, args, kwargs = salt.utils.args.parse_function(
        "amod.afunc[str1, str2, kw1=val1, kw2=val2]"
    )
    assert fun is None
    assert args is None
    assert kwargs is None


def test_parse_function_brackets_unbalanced():
    fun, args, kwargs = salt.utils.args.parse_function(
        "amod.afunc(str1, str2, kw1=val1, kw2=val2"
    )
    assert fun is None
    assert args is None
    assert kwargs is None
    fun, args, kwargs = salt.utils.args.parse_function(
        "amod.afunc(str1, str2, kw1=val1, kw2=val2]"
    )
    assert fun is None
    assert args is None
    assert kwargs is None
    fun, args, kwargs = salt.utils.args.parse_function(
        "amod.afunc(str1, str2, kw1=(val1[val2)], kw2=val2)"
    )
    assert fun is None
    assert args is None
    assert kwargs is None


def test_parse_function_brackets_in_quotes():
    fun, args, kwargs = salt.utils.args.parse_function(
        'amod.afunc(str1, str2, kw1="(val1[val2)]", kw2=val2)'
    )
    assert fun == "amod.afunc"
    assert args == ["str1", "str2"]
    assert kwargs == {"kw1": "(val1[val2)]", "kw2": "val2"}


def test_parse_function_quotes():
    fun, args, kwargs = salt.utils.args.parse_function(
        "amod.afunc(\"double \\\" single '\", 'double \" single \\'',"
        ' kw1="equal=equal", kw2=val2)'
    )
    assert fun == "amod.afunc"
    assert args == ["double \" single '", "double \" single '"]
    assert kwargs == {"kw1": "equal=equal", "kw2": "val2"}


def test_yamlify_arg():
    """
    Test that we properly yamlify CLI input. In several of the tests below `is`
    is used instead of ==. This is because we want to confirm that the return
    value is not a copy of the original, but the same instance as the original.
    """

    def _yamlify_arg(item):
        log.debug("Testing yamlify_arg with %r", item)
        return salt.utils.args.yamlify_arg(item)

    # Make sure non-strings are just returned back
    for item in (True, False, None, 123, 45.67, ["foo"], {"foo": "bar"}):
        assert _yamlify_arg(item) is item

    # Make sure whitespace-only isn't loaded as None
    for item in ("", "\t", " "):
        assert _yamlify_arg(item) is item

    # This value would be loaded as an int (123), the underscores would be
    # ignored. Test that we identify this case and return the original
    # value.
    item = "1_2_3"
    assert _yamlify_arg(item) is item

    # The '#' is treated as a comment when not part of a data structure, we
    # don't want that behavior
    for item in ("# hash at beginning", "Hello world! # hash elsewhere"):
        assert _yamlify_arg(item) is item

    # However we _do_ want the # to be intact if it _is_ within a data
    # structure.
    item = '["foo", "bar", "###"]'
    assert _yamlify_arg(item) == ["foo", "bar", "###"]
    item = '{"foo": "###"}'
    assert _yamlify_arg(item) == {"foo": "###"}

    # The string "None" should load _as_ None
    assert _yamlify_arg("None") is None

    # Leading dashes, or strings containing colons, will result in lists
    # and dicts, and we only want to load lists and dicts when the strings
    # look like data structures.
    for item in ("- foo", "foo: bar"):
        assert _yamlify_arg(item) is item

    # Make sure we don't load '|' as ''
    item = "|"
    assert _yamlify_arg(item) is item

    # Make sure we don't load '!' as something else (None in 2018.3, '' in newer)
    item = "!"
    assert _yamlify_arg(item) is item

    # Make sure we load ints, floats, and strings correctly
    assert _yamlify_arg("123") == 123
    assert _yamlify_arg("45.67") == 45.67
    assert _yamlify_arg("foo") == "foo"

    # We tested list/dict loading above, but there is separate logic when
    # the string contains a '#', so we need to test again here.
    assert _yamlify_arg('["foo", "bar"]') == ["foo", "bar"]
    assert _yamlify_arg('{"foo": "bar"}') == {"foo": "bar"}

    # Make sure that an empty string is loaded properly.
    assert _yamlify_arg("   ") == "   "

    # Make sure that we don't improperly load strings that would be
    # interpreted by PyYAML as YAML document start/end.
    assert _yamlify_arg("---") == "---"
    assert _yamlify_arg("--- ") == "--- "
    assert _yamlify_arg("...") == "..."
    assert _yamlify_arg(" ...") == " ..."

    # Make sure that non-printable whitespace is not YAML-loaded
    assert _yamlify_arg("foo\t\nbar") == "foo\t\nbar"


def test_arguments_regex():
    argument_matches = (
        ("pip=1.1", ("pip", "1.1")),
        ("pip==1.1", None),
        ("pip=1.2=1", ("pip", "1.2=1")),
    )
    for argument, match in argument_matches:
        if match is None:
            assert salt.utils.args.KWARG_REGEX.match(argument) is None
        else:
            assert salt.utils.args.KWARG_REGEX.match(argument).groups() == match



"""
.. versionadded:: 2016.3.0

This is a queue with postgres as the backend.  It uses the jsonb store to
store information for queues.

:depends:       python-psycopg2

To enable this queue, the following needs to be configured in your master
config. These are the defaults:

.. code-block:: yaml

    queue.pgjsonb.host: 'salt'
    queue.pgjsonb.user: 'salt'
    queue.pgjsonb.password: 'salt'
    queue.pgjsonb.dbname: 'salt'
    queue.pgjsonb.port: 5432

Use the following Pg database schema:

.. code-block:: sql

    CREATE DATABASE  salt WITH ENCODING 'utf-8';

    --
    -- Table structure for table `salt`
    --
    DROP TABLE IF EXISTS salt;
    CREATE OR REPLACE TABLE salt(
       id SERIAL PRIMARY KEY,
       data jsonb NOT NULL
    );

.. code-block:: bash

    salt-run queue.insert test '{"name": "redis", "host": "172.16.0.8", "port": 6379}' backend=pgjsonb
    salt-run queue.process_queue test all backend=pgjsonb
"""

import logging
import sys
from contextlib import contextmanager

import salt.utils.json
from salt.exceptions import SaltInvocationError, SaltMasterError

try:
    import psycopg2

    HAS_PG = True
except ImportError:
    HAS_PG = False


log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pgjsonb"


def __virtual__():
    if HAS_PG is False:
        return False
    return __virtualname__


@contextmanager
def _conn(commit=False):
    """
    Return an postgres cursor
    """
    defaults = {
        "host": "localhost",
        "user": "salt",
        "password": "salt",
        "dbname": "salt",
        "port": 5432,
    }

    conn_kwargs = {}
    for key, value in defaults.items():
        conn_kwargs[key] = __opts__.get(f"queue.{__virtualname__}.{key}", value)
    try:
        conn = psycopg2.connect(**conn_kwargs)
    except psycopg2.OperationalError as exc:
        raise SaltMasterError(f"pgjsonb returner could not connect to database: {exc}")

    cursor = conn.cursor()

    try:
        yield cursor
    except psycopg2.DatabaseError as err:
        error = err.args
        sys.stderr.write(str(error))
        cursor.execute("ROLLBACK")
        raise
    else:
        if commit:
            cursor.execute("COMMIT")
        else:
            cursor.execute("ROLLBACK")
    finally:
        conn.close()


def _list_tables(cur):
    cmd = "select relname from pg_class where relkind='r' and relname !~ '^(pg_|sql_)';"
    log.debug("SQL Query: %s", cmd)
    cur.execute(cmd)
    result = cur.fetchall()
    return [x[0] for x in result]


def _create_table(cur, queue):
    cmd = f"CREATE TABLE {queue}(id SERIAL PRIMARY KEY, data jsonb NOT NULL)"
    log.debug("SQL Query: %s", cmd)
    cur.execute(cmd)
    return True


def _list_items(queue):
    """
    Private function to list contents of a queue
    """
    with _conn() as cur:
        cmd = f"SELECT data FROM {queue}"
        log.debug("SQL Query: %s", cmd)
        cur.execute(cmd)
        contents = cur.fetchall()
        return contents


def list_queues():
    """
    Return a list of Salt Queues on the Salt Master
    """
    with _conn() as cur:
        queues = _list_tables(cur)
    return queues


def list_items(queue):
    """
    List contents of a queue
    """
    itemstuple = _list_items(queue)
    items = [item[0] for item in itemstuple]
    return items


def list_length(queue):
    """
    Provide the number of items in a queue
    """
    items = _list_items(queue)
    return len(items)


def _queue_exists(queue):
    """
    Does this queue exist
    :param queue: Name of the queue
    :type str
    :return: True if this queue exists and
    False otherwise
    :rtype bool
    """
    return queue in list_queues()


def handle_queue_creation(queue):
    if not _queue_exists(queue):
        with _conn(commit=True) as cur:
            log.debug("Queue %s does not exist. Creating", queue)
            _create_table(cur, queue)
    else:
        log.debug("Queue %s already exists.", queue)


def insert(queue, items):
    """
    Add an item or items to a queue
    """
    handle_queue_creation(queue)

    with _conn(commit=True) as cur:
        if isinstance(items, dict):
            items = salt.utils.json.dumps(items)
            cmd = f"INSERT INTO {queue}(data) VALUES('{items}')"
            log.debug("SQL Query: %s", cmd)
            try:
                cur.execute(cmd)
            except psycopg2.IntegrityError as esc:
                return "Item already exists in this queue. postgres error: {}".format(
                    esc
                )
        if isinstance(items, list):
            items = [(salt.utils.json.dumps(el),) for el in items]
            cmd = f"INSERT INTO {queue}(data) VALUES (%s)"
            log.debug("SQL Query: %s", cmd)
            try:
                cur.executemany(cmd, items)
            except psycopg2.IntegrityError as esc:
                return (
                    "One or more items already exists in this queue. "
                    "postgres error: {}".format(esc)
                )
    return True


def delete(queue, items):
    """
    Delete an item or items from a queue
    """
    with _conn(commit=True) as cur:
        if isinstance(items, dict):
            cmd = "DELETE FROM {} WHERE data = '{}'".format(
                queue, salt.utils.json.dumps(items)
            )
            log.debug("SQL Query: %s", cmd)
            cur.execute(cmd)
            return True
        if isinstance(items, list):
            items = [(salt.utils.json.dumps(el),) for el in items]
            cmd = f"DELETE FROM {queue} WHERE data = %s"
            log.debug("SQL Query: %s", cmd)
            cur.executemany(cmd, items)
    return True


def pop(queue, quantity=1, is_runner=False):
    """
    Pop one or more or all items from the queue return them.
    """
    cmd = f"SELECT id, data FROM {queue}"
    if quantity != "all":
        try:
            quantity = int(quantity)
        except ValueError as exc:
            error_txt = 'Quantity must be an integer or "all".\nError: "{}".'.format(
                exc
            )
            raise SaltInvocationError(error_txt)
        cmd = "".join([cmd, f" LIMIT {quantity};"])
    log.debug("SQL Query: %s", cmd)
    items = []
    with _conn(commit=True) as cur:
        cur.execute(cmd)
        result = cur.fetchall()
        if len(result) > 0:
            ids = [str(item[0]) for item in result]
            items = [item[1] for item in result]
            idlist = "','".join(ids)
            del_cmd = f"DELETE FROM {queue} WHERE id IN ('{idlist}');"

            log.debug("SQL Query: %s", del_cmd)

            cur.execute(del_cmd)
    return items



import pytest

from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def chocolatey(modules):
    return modules.chocolatey


@pytest.fixture()
def clean(chocolatey):
    result = chocolatey.unbootstrap()

    # Try to get the new version, should throw an error
    try:
        chocolatey_version = chocolatey.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None

    # Assert the chocolatey is not installed
    assert chocolatey_version is None
    try:
        # We're yielding "result" here so we can see any problems with
        # unbootstrap if the test fails
        yield result
    finally:
        try:
            # If chocolatey is not installed, this will throw an error
            chocolatey.chocolatey_version(refresh=True)
            # If we get this far, chocolatey is installed... let's uninstall
            chocolatey.unbootstrap()
        except CommandExecutionError:
            pass


def test_bootstrap(chocolatey, clean):
    # We're defining "result" here to see the output of the bootstrap function
    # if the test fails
    result = chocolatey.bootstrap()
    # Let's run it outside the try/except to see what the error is
    try:
        chocolatey_version = chocolatey.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None
    assert chocolatey_version is not None


def test_bootstrap_version(chocolatey, clean):
    chocolatey.bootstrap(version="1.4.0")
    try:
        chocolatey_version = chocolatey.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None
    assert chocolatey_version == "1.4.0"



import io

import pytest

import salt.proxy.junos as junos
from tests.support.mock import ANY, patch

try:
    import jxmlease  # pylint: disable=unused-import
    from jnpr.junos.device import Device  # pylint: disable=unused-import
    from jnpr.junos.exception import ConnectError

    HAS_JUNOS = True
except ImportError:
    HAS_JUNOS = False


pytestmark = [
    pytest.mark.skipif(
        not HAS_JUNOS, reason="The junos-eznc and jxmlease modules are required"
    ),
]


@pytest.fixture
def opts():
    return {
        "proxy": {
            "username": "xxxx",
            "password]": "xxx",
            "host": "junos",
            "port": "960",
        }
    }


@pytest.fixture
def configure_loader_modules():
    return {junos: {"__pillar__": {}}}


def test_init(opts):
    with patch("ncclient.manager.connect") as mock_connect:
        junos.init(opts)
        assert junos.thisproxy.get("initialized")
        mock_connect.assert_called_with(
            allow_agent=True,
            device_params={"name": "junos", "local": False, "use_filter": False},
            host="junos",
            hostkey_verify=False,
            key_filename=None,
            look_for_keys=True,
            password=None,
            port="960",
            sock_fd=None,
            ssh_config=ANY,
            timeout=30,
            username="xxxx",
        )


def test_init_err(opts):
    with patch("ncclient.manager.connect") as mock_connect:
        mock_connect.side_effect = ConnectError
        junos.init(opts)
        assert not junos.thisproxy.get("initialized")


def test_alive(opts):
    with patch("ncclient.manager.connect") as mock_connect:
        junos.init(opts)
        junos.thisproxy["conn"]._conn._session._buffer = io.BytesIO()
        assert junos.alive(opts)
        assert junos.thisproxy.get("initialized")



"""
Make api awesomeness
"""

import copy
import inspect
import logging
import os

import salt.auth
import salt.client
import salt.client.ssh.client
import salt.config
import salt.daemons.masterapi
import salt.exceptions
import salt.runner
import salt.syspaths
import salt.utils.args
import salt.utils.minions
import salt.wheel
from salt.defaults import DEFAULT_TARGET_DELIM

log = logging.getLogger(__name__)


def sorted_permissions(perms):
    """
    Return a sorted list of the passed in permissions, de-duplicating in the process
    """
    _str_perms = []
    _non_str_perms = []
    for entry in perms:
        if isinstance(entry, str):
            if entry in _str_perms:
                continue
            _str_perms.append(entry)
            continue
        if entry in _non_str_perms:
            continue
        _non_str_perms.append(entry)
    return sorted(_str_perms) + sorted(_non_str_perms, key=repr)


def sum_permissions(token, eauth):
    """
    Returns the sum of '*', user-specific and group specific permissions
    """
    perms = eauth.get(token["name"], [])
    perms.extend(eauth.get("*", []))

    if "groups" in token and token["groups"]:
        user_groups = set(token["groups"])
        eauth_groups = {i.rstrip("%") for i in eauth.keys() if i.endswith("%")}

        for group in user_groups & eauth_groups:
            perms.extend(eauth[f"{group}%"])
    return perms


class NetapiClient:
    """
    Provide a uniform method of accessing the various client interfaces in Salt
    in the form of low-data data structures. For example:

    >>> client = NetapiClient(__opts__)
    >>> lowstate = {'client': 'local', 'tgt': '*', 'fun': 'test.ping', 'arg': ''}
    >>> client.run(lowstate)
    """

    def __init__(self, opts):
        self.opts = opts
        apiopts = copy.deepcopy(self.opts)
        apiopts["enable_ssh_minions"] = True
        apiopts["cachedir"] = os.path.join(opts["cachedir"], "saltapi")
        if not os.path.exists(apiopts["cachedir"]):
            os.makedirs(apiopts["cachedir"])
        self.resolver = salt.auth.Resolver(apiopts)
        self.loadauth = salt.auth.LoadAuth(apiopts)
        self.key = salt.daemons.masterapi.access_keys(apiopts)
        self.ckminions = salt.utils.minions.CkMinions(apiopts)

    def _is_master_running(self):
        """
        Perform a lightweight check to see if the master daemon is running

        Note, this will return an invalid success if the master crashed or was
        not shut down cleanly.
        """
        # Windows doesn't have IPC. Assume the master is running.
        # At worse, it will error 500.
        if salt.utils.platform.is_windows():
            return True

        if self.opts["transport"] == "tcp":
            ipc_file = "publish_pull.ipc"
        else:
            ipc_file = "workers.ipc"
        return os.path.exists(os.path.join(self.opts["sock_dir"], ipc_file))

    def _prep_auth_info(self, clear_load):
        sensitive_load_keys = []
        key = None
        if "token" in clear_load:
            auth_type = "token"
            err_name = "TokenAuthenticationError"
            sensitive_load_keys = ["token"]
            return auth_type, err_name, key, sensitive_load_keys
        elif "eauth" in clear_load:
            auth_type = "eauth"
            err_name = "EauthAuthenticationError"
            sensitive_load_keys = ["username", "password"]
            return auth_type, err_name, key, sensitive_load_keys
        raise salt.exceptions.EauthAuthenticationError(
            "No authentication credentials given"
        )

    def _authorize_ssh(self, low):
        auth_type, err_name, key, sensitive_load_keys = self._prep_auth_info(low)
        auth_check = self.loadauth.check_authentication(low, auth_type, key=key)
        auth_list = auth_check.get("auth_list", [])
        error = auth_check.get("error")
        if error:
            raise salt.exceptions.EauthAuthenticationError(error)
        delimiter = low.get("kwargs", {}).get("delimiter", DEFAULT_TARGET_DELIM)
        _res = self.ckminions.check_minions(
            low["tgt"], low.get("tgt_type", "glob"), delimiter
        )
        minions = _res.get("minions", list())
        missing = _res.get("missing", list())
        authorized = self.ckminions.auth_check(
            auth_list,
            low["fun"],
            low.get("arg", []),
            low["tgt"],
            low.get("tgt_type", "glob"),
            minions=minions,
        )
        if not authorized:
            raise salt.exceptions.EauthAuthenticationError(
                "Authorization error occurred."
            )

    def run(self, low):
        """
        Execute the specified function in the specified client by passing the
        lowstate
        """
        # Eauth currently requires a running daemon and commands run through
        # this method require eauth so perform a quick check to raise a
        # more meaningful error.
        if not self._is_master_running():
            raise salt.exceptions.SaltDaemonNotRunning("Salt Master is not available.")

        if low.get("client") not in CLIENTS:
            raise salt.exceptions.SaltInvocationError(
                "Invalid client specified: '{}'".format(low.get("client"))
            )

        if low.get("client") not in self.opts.get("netapi_enable_clients"):
            raise salt.exceptions.SaltInvocationError(
                "Client disabled: '{}'. Add to 'netapi_enable_clients' master config option to enable.".format(
                    low.get("client")
                )
            )

        if not ("token" in low or "eauth" in low):
            raise salt.exceptions.EauthAuthenticationError(
                "No authentication credentials given"
            )

        if low.get("raw_shell") and not self.opts.get("netapi_allow_raw_shell"):
            raise salt.exceptions.EauthAuthenticationError(
                "Raw shell option not allowed."
            )

        if low["client"] == "ssh":
            self._authorize_ssh(low)

        l_fun = getattr(self, low["client"])
        f_call = salt.utils.args.format_call(l_fun, low)
        return l_fun(*f_call.get("args", ()), **f_call.get("kwargs", {}))

    def local_async(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` asynchronously

        Wraps :py:meth:`salt.client.LocalClient.run_job`.

        :return: job ID
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.run_job(*args, **kwargs)

    def local(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` synchronously

        See :py:meth:`salt.client.LocalClient.cmd` for all available
        parameters.

        Sends a command from the master to the targeted minions. This is the
        same interface that Salt's own CLI uses. Note the ``arg`` and ``kwarg``
        parameters are sent down to the minion(s) and the given function,
        ``fun``, is called with those parameters.

        :return: Returns the result from the execution module
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.cmd(*args, **kwargs)

    def local_subset(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` against subsets of minions

        .. versionadded:: 2016.3.0

        Wraps :py:meth:`salt.client.LocalClient.cmd_subset`
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.cmd_subset(*args, **kwargs)

    def local_batch(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` against batches of minions

        .. versionadded:: 0.8.4

        Wraps :py:meth:`salt.client.LocalClient.cmd_batch`

        :return: Returns the result from the exeuction module for each batch of
            returns
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.cmd_batch(*args, **kwargs)

    def ssh(self, *args, **kwargs):
        """
        Run salt-ssh commands synchronously

        Wraps :py:meth:`salt.client.ssh.client.SSHClient.cmd_sync`.

        :return: Returns the result from the salt-ssh command
        """
        with salt.client.ssh.client.SSHClient(
            mopts=self.opts, disable_custom_roster=True
        ) as client:
            return client.cmd_sync(kwargs)

    def runner(self, fun, timeout=None, full_return=False, **kwargs):
        """
        Run `runner modules <all-salt.runners>` synchronously

        Wraps :py:meth:`salt.runner.RunnerClient.cmd_sync`.

        Note that runner functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: Returns the result from the runner module
        """
        kwargs["fun"] = fun
        runner = salt.runner.RunnerClient(self.opts)
        return runner.cmd_sync(kwargs, timeout=timeout, full_return=full_return)

    def runner_async(self, fun, **kwargs):
        """
        Run `runner modules <all-salt.runners>` asynchronously

        Wraps :py:meth:`salt.runner.RunnerClient.cmd_async`.

        Note that runner functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: event data and a job ID for the executed function.
        """
        kwargs["fun"] = fun
        runner = salt.runner.RunnerClient(self.opts)
        return runner.cmd_async(kwargs)

    def wheel(self, fun, **kwargs):
        """
        Run :ref:`wheel modules <all-salt.wheel>` synchronously

        Wraps :py:meth:`salt.wheel.WheelClient.master_call`.

        Note that wheel functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: Returns the result from the wheel module
        """
        kwargs["fun"] = fun
        wheel = salt.wheel.WheelClient(self.opts)
        return wheel.cmd_sync(kwargs)

    def wheel_async(self, fun, **kwargs):
        """
        Run :ref:`wheel modules <all-salt.wheel>` asynchronously

        Wraps :py:meth:`salt.wheel.WheelClient.master_call`.

        Note that wheel functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: Returns the result from the wheel module
        """
        kwargs["fun"] = fun
        wheel = salt.wheel.WheelClient(self.opts)
        return wheel.cmd_async(kwargs)


CLIENTS = [
    name
    for name, _ in inspect.getmembers(NetapiClient, predicate=None)
    if not (name == "run" or name.startswith("_"))
]



"""
Manage PagerDuty schedules.

Example:

.. code-block:: yaml

    ensure test schedule:
        pagerduty_schedule.present:
            - name: 'bruce test schedule level1'
            - schedule:
                name: 'bruce test schedule level1'
                time_zone: 'Pacific Time (US & Canada)'
                schedule_layers:
                    - name: 'Schedule Layer 1'
                      start: '2015-01-01T00:00:00'
                      users:
                        - user:
                            'id': 'Bruce TestUser1'
                          member_order: 1
                        - user:
                            'id': 'Bruce TestUser2'
                          member_order: 2
                        - user:
                            'id': 'bruce+test3@lyft.com'
                          member_order: 3
                        - user:
                            'id': 'bruce+test4@lyft.com'
                          member_order: 4
                      rotation_virtual_start: '2015-01-01T00:00:00'
                      priority: 1
                      rotation_turn_length_seconds: 604800

"""


def __virtual__():
    """
    Only load if the pygerduty module is available in __salt__
    """
    if "pagerduty_util.get_resource" in __salt__:
        return "pagerduty_schedule"
    return (False, "pagerduty_util module could not be loaded")


def present(profile="pagerduty", subdomain=None, api_key=None, **kwargs):
    """
    Ensure that a pagerduty schedule exists.
    This method accepts as args everything defined in
    https://developer.pagerduty.com/documentation/rest/schedules/create.
    This means that most arguments are in a dict called "schedule."

    User id's can be pagerduty id, or name, or email address.
    """
    # for convenience, we accept id, name, or email as the user id.
    kwargs["schedule"]["name"] = kwargs["name"]  # match PD API structure
    for schedule_layer in kwargs["schedule"]["schedule_layers"]:
        for user in schedule_layer["users"]:
            u = __salt__["pagerduty_util.get_resource"](
                "users",
                user["user"]["id"],
                ["email", "name", "id"],
                profile=profile,
                subdomain=subdomain,
                api_key=api_key,
            )
            if u is None:
                raise Exception(f"unknown user: {user}")
            user["user"]["id"] = u["id"]
    r = __salt__["pagerduty_util.resource_present"](
        "schedules", ["name", "id"], _diff, profile, subdomain, api_key, **kwargs
    )
    return r


def absent(profile="pagerduty", subdomain=None, api_key=None, **kwargs):
    """
    Ensure that a pagerduty schedule does not exist.
    Name can be pagerduty schedule id or pagerduty schedule name.
    """
    r = __salt__["pagerduty_util.resource_absent"](
        "schedules", ["name", "id"], profile, subdomain, api_key, **kwargs
    )
    return r


def _diff(state_data, resource_object):
    """helper method to compare salt state info with the PagerDuty API json structure,
    and determine if we need to update.

    returns the dict to pass to the PD API to perform the update, or empty dict if no update.
    """

    state_data["id"] = resource_object["schedule"]["id"]
    objects_differ = None

    # first check all the easy top-level properties: everything except the schedule_layers.
    for k, v in state_data["schedule"].items():
        if k == "schedule_layers":
            continue
        if v != resource_object["schedule"][k]:
            objects_differ = "{} {} {}".format(k, v, resource_object["schedule"][k])
            break

    # check schedule_layers
    if not objects_differ:
        for layer in state_data["schedule"]["schedule_layers"]:
            # find matching layer name
            resource_layer = None
            for resource_layer in resource_object["schedule"]["schedule_layers"]:
                found = False
                if layer["name"] == resource_layer["name"]:
                    found = True
                    break
            if not found:
                objects_differ = "layer {} missing".format(layer["name"])
                break
            # set the id, so that we will update this layer instead of creating a new one
            layer["id"] = resource_layer["id"]
            # compare contents of layer and resource_layer
            for k, v in layer.items():
                if k == "users":
                    continue
                if k == "start":
                    continue
                if v != resource_layer[k]:
                    objects_differ = "layer {} key {} {} != {}".format(
                        layer["name"], k, v, resource_layer[k]
                    )
                    break
            if objects_differ:
                break
            # compare layer['users']
            if len(layer["users"]) != len(resource_layer["users"]):
                objects_differ = "num users in layer {} {} != {}".format(
                    layer["name"], len(layer["users"]), len(resource_layer["users"])
                )
                break

            for user1 in layer["users"]:
                found = False
                user2 = None
                for user2 in resource_layer["users"]:
                    # deal with PD API bug: when you submit member_order=N, you get back member_order=N+1
                    if user1["member_order"] == user2["member_order"] - 1:
                        found = True
                        break
                if not found:
                    objects_differ = "layer {} no one with member_order {}".format(
                        layer["name"], user1["member_order"]
                    )
                    break
                if user1["user"]["id"] != user2["user"]["id"]:
                    objects_differ = "layer {} user at member_order {} {} != {}".format(
                        layer["name"],
                        user1["member_order"],
                        user1["user"]["id"],
                        user2["user"]["id"],
                    )
                    break
    if objects_differ:
        return state_data
    else:
        return {}



"""
Decorator and functions to profile Salt using cProfile
"""

import datetime
import logging
import os
import pstats
import subprocess

import salt.utils.files
import salt.utils.hashutils
import salt.utils.path
import salt.utils.stringutils

log = logging.getLogger(__name__)

try:
    import cProfile

    HAS_CPROFILE = True
except ImportError:
    HAS_CPROFILE = False


def profile_func(filename=None):
    """
    Decorator for adding profiling to a nested function in Salt
    """

    def proffunc(fun):
        def profiled_func(*args, **kwargs):
            logging.info("Profiling function %s", fun.__name__)
            try:
                profiler = cProfile.Profile()
                retval = profiler.runcall(fun, *args, **kwargs)
                profiler.dump_stats(filename or f"{fun.__name__}_func.profile")
            except OSError:
                logging.exception("Could not open profile file %s", filename)

            return retval

        return profiled_func

    return proffunc


def activate_profile(test=True):
    pr = None
    if test:
        if HAS_CPROFILE:
            pr = cProfile.Profile()
            pr.enable()
        else:
            log.error("cProfile is not available on your platform")
    return pr


def output_profile(pr, stats_path="/tmp/stats", stop=False, id_=None):
    if pr is not None and HAS_CPROFILE:
        try:
            pr.disable()
            if not os.path.isdir(stats_path):
                os.makedirs(stats_path)
            date = datetime.datetime.now().isoformat()
            if id_ is None:
                id_ = salt.utils.hashutils.random_hash(size=32)
            ficp = os.path.join(stats_path, f"{id_}.{date}.pstats")
            fico = os.path.join(stats_path, f"{id_}.{date}.dot")
            ficn = os.path.join(stats_path, f"{id_}.{date}.stats")
            if not os.path.exists(ficp):
                pr.dump_stats(ficp)
                with salt.utils.files.fopen(ficn, "w") as fic:
                    pstats.Stats(pr, stream=fic).sort_stats("cumulative")
            log.info("PROFILING: %s generated", ficp)
            log.info("PROFILING (cumulative): %s generated", ficn)
            pyprof = salt.utils.path.which("pyprof2calltree")
            cmd = [pyprof, "-i", ficp, "-o", fico]
            if pyprof:
                failed = False
                try:
                    pro = subprocess.Popen(
                        cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    )
                except OSError:
                    failed = True
                if pro.returncode:
                    failed = True
                if failed:
                    log.error("PROFILING (dot problem")
                else:
                    log.info("PROFILING (dot): %s generated", fico)
                log.trace("pyprof2calltree output:")
                log.trace(
                    salt.utils.stringutils.to_str(pro.stdout.read()).strip()
                    + salt.utils.stringutils.to_str(pro.stderr.read()).strip()
                )
            else:
                log.info("You can run %s for additional stats.", cmd)
        finally:
            if not stop:
                pr.enable()
    return pr



import sys
import time

import pytest


@pytest.fixture(scope="module")
def pkg_name(salt_call_cli, grains):
    if sys.platform.startswith("win"):
        ret = salt_call_cli.run("--local", "winrepo.update_git_repos")
        assert ret.returncode == 0
        attempts = 3
        while attempts:
            attempts -= 1
            ret = salt_call_cli.run("--local", "pkg.refresh_db")
            if ret.returncode:
                time.sleep(5)
                continue
            break
        else:
            pytest.fail("Failed to run 'pkg.refresh_db' 3 times.")
        return "putty"
    elif grains["os_family"] == "RedHat":
        if grains["os"] == "VMware Photon OS":
            return "snoopy"
        elif grains["osfinger"] == "Amazon Linux-2023":
            return "dnf-utils"
        return "units"
    elif grains["os_family"] == "Debian":
        return "ifenslave"
    return "figlet"


def test_pkg_install(salt_call_cli, pkg_name):
    ret = salt_call_cli.run("--local", "state.single", "pkg.installed", pkg_name)
    assert ret.returncode == 0



"""
Wrapper for rsync

.. versionadded:: 2014.1.0

This data can also be passed into :ref:`pillar <pillar-walk-through>`.
Options passed into opts will overwrite options passed into pillar.
"""

import errno
import logging
import re
import tempfile

import salt.utils.files
import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

__virtualname__ = "rsync"


def __virtual__():
    """
    Only load module if rsync binary is present
    """
    if salt.utils.path.which("rsync"):
        return __virtualname__
    return (
        False,
        "The rsync execution module cannot be loaded: "
        "the rsync binary is not in the path.",
    )


def _check(delete, force, update, passwordfile, exclude, excludefrom, dryrun, rsh):
    """
    Generate rsync options
    """
    options = ["-avz"]

    if delete:
        options.append("--delete")
    if force:
        options.append("--force")
    if update:
        options.append("--update")
    if rsh:
        options.append(f"--rsh={rsh}")
    if passwordfile:
        options.extend(["--password-file", passwordfile])
    if excludefrom:
        options.extend(["--exclude-from", excludefrom])
        if exclude:
            exclude = False
    if exclude:
        if isinstance(exclude, list):
            for ex_ in exclude:
                options.extend(["--exclude", ex_])
        else:
            options.extend(["--exclude", exclude])
    if dryrun:
        options.append("--dry-run")
    return options


def rsync(
    src,
    dst,
    delete=False,
    force=False,
    update=False,
    passwordfile=None,
    exclude=None,
    excludefrom=None,
    dryrun=False,
    rsh=None,
    additional_opts=None,
    saltenv="base",
):
    """
    .. versionchanged:: 2016.3.0
        Return data now contains just the output of the rsync command, instead
        of a dictionary as returned from :py:func:`cmd.run_all
        <salt.modules.cmdmod.run_all>`.

    Rsync files from src to dst

    src
        The source location where files will be rsynced from.

    dst
        The destination location where files will be rsynced to.

    delete : False
        Whether to enable the rsync `--delete` flag, which
        will delete extraneous files from dest dirs

    force : False
        Whether to enable the rsync `--force` flag, which
        will force deletion of dirs even if not empty.

    update : False
        Whether to enable the rsync `--update` flag, which
        forces rsync to skip any files which exist on the
        destination and have a modified time that is newer
        than the source file.

    passwordfile
        A file that contains a password for accessing an
        rsync daemon.  The file should contain just the
        password.

    exclude
        Whether to enable the rsync `--exclude` flag, which
        will exclude files matching a PATTERN.

    excludefrom
        Whether to enable the rsync `--excludefrom` flag, which
        will read exclude patterns from a file.

    dryrun : False
        Whether to enable the rsync `--dry-run` flag, which
        will perform a trial run with no changes made.

    rsh
        Whether to enable the rsync `--rsh` flag, to
        specify the remote shell to use.

    additional_opts
        Any additional rsync options, should be specified as a list.

    saltenv
        Specify a salt fileserver environment to be used.

    CLI Example:

    .. code-block:: bash

        salt '*' rsync.rsync /path/to/src /path/to/dest delete=True update=True passwordfile=/etc/pass.crt exclude=exclude/dir
        salt '*' rsync.rsync /path/to/src delete=True excludefrom=/xx.ini
        salt '*' rsync.rsync /path/to/src delete=True exclude='[exclude1/dir,exclude2/dir]' additional_opts='["--partial", "--bwlimit=5000"]'
    """
    if not src:
        src = __salt__["config.option"]("rsync.src")
    if not dst:
        dst = __salt__["config.option"]("rsync.dst")
    if not delete:
        delete = __salt__["config.option"]("rsync.delete")
    if not force:
        force = __salt__["config.option"]("rsync.force")
    if not update:
        update = __salt__["config.option"]("rsync.update")
    if not passwordfile:
        passwordfile = __salt__["config.option"]("rsync.passwordfile")
    if not exclude:
        exclude = __salt__["config.option"]("rsync.exclude")
    if not excludefrom:
        excludefrom = __salt__["config.option"]("rsync.excludefrom")
    if not dryrun:
        dryrun = __salt__["config.option"]("rsync.dryrun")
    if not rsh:
        rsh = __salt__["config.option"]("rsync.rsh")
    if not src or not dst:
        raise SaltInvocationError("src and dst cannot be empty")

    tmp_src = None
    if src.startswith("salt://"):
        _src = src
        _path = re.sub("salt://", "", _src)
        src_is_dir = False
        if _path in __salt__["cp.list_master_dirs"](saltenv=saltenv):
            src_is_dir = True

        if src_is_dir:
            tmp_src = tempfile.mkdtemp()
            dir_src = __salt__["cp.get_dir"](_src, tmp_src, saltenv)
            if dir_src:
                src = tmp_src
                # Ensure src ends in / so we
                # get the contents not the tmpdir
                # itself.
                if not src.endswith("/"):
                    src = f"{src}/"
            else:
                raise CommandExecutionError(f"{src} does not exist")
        else:
            tmp_src = salt.utils.files.mkstemp()
            file_src = __salt__["cp.get_file"](_src, tmp_src, saltenv)
            if file_src:
                src = tmp_src
            else:
                raise CommandExecutionError(f"{src} does not exist")

    option = _check(
        delete, force, update, passwordfile, exclude, excludefrom, dryrun, rsh
    )

    if additional_opts and isinstance(additional_opts, list):
        option = option + additional_opts

    cmd = ["rsync"] + option + [src, dst]
    log.debug("Running rsync command: %s", cmd)
    try:
        return __salt__["cmd.run_all"](cmd, python_shell=False)
    except OSError as exc:
        raise CommandExecutionError(exc.strerror)
    finally:
        if tmp_src:
            __salt__["file.remove"](tmp_src)


def version():
    """
    .. versionchanged:: 2016.3.0
        Return data now contains just the version number as a string, instead
        of a dictionary as returned from :py:func:`cmd.run_all
        <salt.modules.cmdmod.run_all>`.

    Returns rsync version

    CLI Example:

    .. code-block:: bash

        salt '*' rsync.version
    """
    try:
        out = __salt__["cmd.run_stdout"](["rsync", "--version"], python_shell=False)
    except OSError as exc:
        raise CommandExecutionError(exc.strerror)
    try:
        return out.split("\n")[0].split()[2]
    except IndexError:
        raise CommandExecutionError("Unable to determine rsync version")


def config(conf_path="/etc/rsyncd.conf"):
    """
    .. versionchanged:: 2016.3.0
        Return data now contains just the contents of the rsyncd.conf as a
        string, instead of a dictionary as returned from :py:func:`cmd.run_all
        <salt.modules.cmdmod.run_all>`.

    Returns the contents of the rsync config file

    conf_path : /etc/rsyncd.conf
        Path to the config file

    CLI Example:

    .. code-block:: bash

        salt '*' rsync.config
    """
    ret = ""
    try:
        with salt.utils.files.fopen(conf_path, "r") as fp_:
            for line in fp_:
                ret += salt.utils.stringutils.to_unicode(line)
    except OSError as exc:
        if exc.errno == errno.ENOENT:
            raise CommandExecutionError(f"{conf_path} does not exist")
        elif exc.errno == errno.EACCES:
            raise CommandExecutionError(f"Unable to read {conf_path}, access denied")
        elif exc.errno == errno.EISDIR:
            raise CommandExecutionError(
                f"Unable to read {conf_path}, path is a directory"
            )
        else:
            raise CommandExecutionError(f"Error {exc.errno}: {exc.strerror}")
    else:
        return ret



"""
Ensure salt.utils.clean_path works with symlinked directories and files
"""

import ctypes

import pytest

import salt.utils.verify


class Symlink:
    """
    symlink(source, link_name) Creates a symbolic link pointing to source named
    link_name
    """

    def __init__(self):
        self._csl = None

    def __call__(self, source, link_name):
        if self._csl is None:
            self._csl = ctypes.windll.kernel32.CreateSymbolicLinkW
            self._csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
            self._csl.restype = ctypes.c_ubyte
        flags = 0
        if source is not None and source.is_dir():
            flags = 1

        if self._csl(str(link_name), str(source), flags) == 0:
            raise ctypes.WinError()


@pytest.fixture(scope="module")
def symlink():
    return Symlink()


@pytest.fixture
def setup_links(tmp_path, symlink):
    to_path = tmp_path / "linkto"
    from_path = tmp_path / "linkfrom"
    if salt.utils.platform.is_windows():
        kwargs = {}
    else:
        kwargs = {"target_is_directory": True}
    if salt.utils.platform.is_windows():
        symlink(to_path, from_path, **kwargs)
    else:
        from_path.symlink_to(to_path, **kwargs)
    return to_path, from_path


def test_clean_path_symlinked_src(setup_links):
    to_path, from_path = setup_links
    test_path = from_path / "test"
    expect_path = str(to_path / "test")
    ret = salt.utils.verify.clean_path(str(from_path), str(test_path))
    assert ret == expect_path, f"{ret} is not {expect_path}"


def test_clean_path_symlinked_tgt(setup_links):
    to_path, from_path = setup_links
    test_path = to_path / "test"
    expect_path = str(to_path / "test")
    ret = salt.utils.verify.clean_path(str(from_path), str(test_path))
    assert ret == expect_path, f"{ret} is not {expect_path}"


def test_clean_path_symlinked_src_unresolved(setup_links):
    to_path, from_path = setup_links
    test_path = from_path / "test"
    expect_path = str(from_path / "test")
    ret = salt.utils.verify.clean_path(str(from_path), str(test_path), realpath=False)
    assert ret == expect_path, f"{ret} is not {expect_path}"
